import numpy as np
import pytest
import sparse

from libertem.io.corrections import CorrectionSet
from libertem.io.corrections.detector import RepairValueError
from libertem.utils.generate import exclude_pixels
from libertem.udf.sum import SumUDF
from libertem.udf.base import NoOpUDF


def _validate(excluded_coords, adjusted, sig_shape):
    for dim in range(len(excluded_coords)):
        excluded_set = frozenset(excluded_coords[dim])
        right_boundaries = set(range(adjusted[dim], sig_shape[dim], adjusted[dim]))
        # The stop at sig_shape[dim]-1 ignores pixels that are on the very right of the sig shape,
        # which are OK
        left_boundaries = set(range(adjusted[dim]-1, sig_shape[dim]-1, adjusted[dim]))
        boundaries = right_boundaries.union(left_boundaries)
        assert excluded_set.isdisjoint(boundaries)


@pytest.mark.parametrize("gain,dark", [
    (np.zeros((128, 128)), np.ones((128, 128))),
    (np.zeros((128, 128)), None),
])
def test_correction_set_zero_gain(lt_ctx, default_raw, gain, dark):
    udf = SumUDF()

    corr = CorrectionSet(dark=dark, gain=gain)
    res = lt_ctx.run_udf(
        dataset=default_raw,
        udf=udf,
        corrections=corr
    )
    assert np.allclose(res['intensity'], 0)


@pytest.mark.parametrize("gain,dark", [
    (np.ones((128, 128)), np.ones((128, 128))),
    (None, np.ones((128, 128))),
])
def test_correction_set_dark_one(lt_ctx, default_raw, default_raw_data, gain, dark):
    udf = SumUDF()

    corr = CorrectionSet(dark=dark, gain=gain)
    res = lt_ctx.run_udf(
        dataset=default_raw,
        udf=udf,
        corrections=corr
    )
    assert np.allclose(res['intensity'], np.sum(default_raw_data - 1, axis=(0, 1)))


def test_patch_pixels(lt_ctx, default_raw, default_raw_data):
    udf = SumUDF()

    # test with empty excluded_pixels array
    corr = CorrectionSet(excluded_pixels=np.array([
        (), ()
    ]).astype(np.int64), gain=np.ones((128, 128)))
    res = lt_ctx.run_udf(
        dataset=default_raw,
        udf=udf,
        corrections=corr
    )
    assert np.allclose(res['intensity'], np.sum(default_raw_data, axis=(0, 1)))


def test_patch_corr_odd(lt_ctx_fast):
    data = np.ones((13, 17, 19, 23, 29, 31))
    excluded_coords = np.array([
            (2, 5), (2, 5), (2, 5)
    ]).astype(np.int64)
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=(23, 29, 31), data=True)

    ds = lt_ctx_fast.load("memory", data=data, sig_dims=3)

    udf = SumUDF()

    corr = CorrectionSet(
        excluded_pixels=excluded_pixels,
        gain=np.ones((23, 29, 31)),
        dark=np.ones((23, 29, 31))
    )
    res = lt_ctx_fast.run_udf(
        dataset=ds,
        udf=udf,
        corrections=corr
    )
    assert np.allclose(res['intensity'], 0)


def test_patch_corr_empty(lt_ctx):
    data = np.ones((13, 17, 19))
    excluded_coords = np.array([
            (1, 2, 3),
    ]).astype(np.int64)
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=(19, ), data=True)

    ds = lt_ctx.load("memory", data=data, sig_dims=1)

    udf = SumUDF()
    with pytest.raises(RepairValueError):
        corr = CorrectionSet(
            excluded_pixels=excluded_pixels,
            gain=np.ones((19, )),
            dark=np.ones((19, ))
        )
    corr = CorrectionSet(
            excluded_pixels=excluded_pixels,
            gain=np.ones((19, )),
            dark=np.ones((19, )),
            allow_empty=True
        )
    res = lt_ctx.run_udf(
        dataset=ds,
        udf=udf,
        corrections=corr
    )
    # The value will be unpatched and remain 0 after gain and dark correction are applied
    assert np.allclose(res['intensity'], 0)


def test_patch_pixels_only_excluded_pixels(lt_ctx, default_raw, default_raw_data):
    udf = SumUDF()
    excluded_pixels = sparse.COO(
        np.zeros((128, 128))
    )
    corr = CorrectionSet(excluded_pixels=excluded_pixels)
    res = lt_ctx.run_udf(
        dataset=default_raw,
        udf=udf,
        corrections=corr
    )
    assert np.allclose(res['intensity'], np.sum(default_raw_data, axis=(0, 1)))


@pytest.mark.with_numba
def test_tileshape_adjustment_1():
    sig_shape = (123, 456)
    tile_shape = (17, 42)
    base_shape = (1, 1)
    excluded_coords = np.array([
        (17, ),
        (42, )
    ])
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=sig_shape, data=True)
    corr = CorrectionSet(excluded_pixels=excluded_pixels)
    adjusted = corr.adjust_tileshape(
        tile_shape=tile_shape, sig_shape=sig_shape, base_shape=base_shape
    )
    assert adjusted == (16, 41)
    _validate(excluded_coords=excluded_coords, adjusted=adjusted, sig_shape=sig_shape)


@pytest.mark.with_numba
def test_tileshape_adjustment_numbacov():
    sig_shape = (123, 456)
    tile_shape = (16, 41)
    base_shape = (3, 3)
    excluded_coords = np.array([
        (17, ),
        (42, )
    ])
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=sig_shape, data=True)
    corr = CorrectionSet(excluded_pixels=excluded_pixels)
    adjusted = corr.adjust_tileshape(
        tile_shape=tile_shape, sig_shape=sig_shape, base_shape=base_shape
    )
    assert adjusted == (15, 39)
    _validate(excluded_coords=excluded_coords, adjusted=adjusted, sig_shape=sig_shape)


def test_tileshape_adjustment_2():
    sig_shape = (123, 456)
    tile_shape = (17, 42)
    base_shape = (1, 1)
    excluded_coords = np.array([
        (17*2 - 1, ),
        (42, )
    ])
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=sig_shape, data=True)
    corr = CorrectionSet(excluded_pixels=excluded_pixels)
    adjusted = corr.adjust_tileshape(
        tile_shape=tile_shape, sig_shape=sig_shape, base_shape=base_shape
    )
    assert adjusted == (16, 41)
    _validate(excluded_coords=excluded_coords, adjusted=adjusted, sig_shape=sig_shape)


def test_tileshape_adjustment_3():
    sig_shape = (123, 456)
    tile_shape = (17, 42)
    base_shape = (1, 1)
    excluded_coords = np.array([
        (122, ),
        (23, )
    ])
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=sig_shape, data=True)
    corr = CorrectionSet(excluded_pixels=excluded_pixels)
    adjusted = corr.adjust_tileshape(
        tile_shape=tile_shape, sig_shape=sig_shape, base_shape=base_shape
    )
    assert adjusted == (17, 42)
    _validate(excluded_coords=excluded_coords, adjusted=adjusted, sig_shape=sig_shape)


def test_tileshape_adjustment_4():
    sig_shape = (123, 456)
    tile_shape = (17, 1)
    base_shape = (1, 1)
    excluded_coords = np.array([
        (122, ),
        (0, )
    ])
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=sig_shape, data=True)
    corr = CorrectionSet(excluded_pixels=excluded_pixels)
    adjusted = corr.adjust_tileshape(
        tile_shape=tile_shape, sig_shape=sig_shape, base_shape=base_shape
    )
    assert adjusted == (17, 2)
    _validate(excluded_coords=excluded_coords, adjusted=adjusted, sig_shape=sig_shape)


def test_tileshape_adjustment_5():
    sig_shape = (123, 1)
    tile_shape = (17, 1)
    base_shape = (1, 1)
    excluded_coords = np.array([
        (122, ),
        (0, )
    ])
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=sig_shape, data=True)
    corr = CorrectionSet(excluded_pixels=excluded_pixels)
    adjusted = corr.adjust_tileshape(
        tile_shape=tile_shape, sig_shape=sig_shape, base_shape=base_shape
    )
    assert adjusted == (17, 1)
    _validate(excluded_coords=excluded_coords, adjusted=adjusted, sig_shape=sig_shape)


def test_tileshape_adjustment_6():
    sig_shape = (123, 456)
    tile_shape = (17, 1)
    base_shape = (1, 1)
    excluded_coords = np.array([
        range(123),
        np.zeros(123, dtype=int)
    ])
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=sig_shape, data=True)
    corr = CorrectionSet(excluded_pixels=excluded_pixels)
    adjusted = corr.adjust_tileshape(
        tile_shape=tile_shape, sig_shape=sig_shape, base_shape=base_shape
    )
    assert adjusted == (123, 2)
    _validate(excluded_coords=excluded_coords, adjusted=adjusted, sig_shape=sig_shape)


def test_tileshape_adjustment_6_1():
    sig_shape = (123, 456)
    tile_shape = (122, 1)
    base_shape = (1, 1)
    excluded_coords = np.array([
        range(123),
        np.zeros(123, dtype=int)
    ])
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=sig_shape, data=True)
    corr = CorrectionSet(excluded_pixels=excluded_pixels)
    adjusted = corr.adjust_tileshape(
        tile_shape=tile_shape, sig_shape=sig_shape, base_shape=base_shape
    )
    print(adjusted)
    assert adjusted == (123, 2)
    _validate(excluded_coords=excluded_coords, adjusted=adjusted, sig_shape=sig_shape)


def test_tileshape_adjustment_6_2():
    sig_shape = (123, 456)
    tile_shape = (1, 1)
    base_shape = (1, 1)
    excluded_coords = np.array([
        range(123),
        np.zeros(123, dtype=int)
    ])
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=sig_shape, data=True)
    corr = CorrectionSet(excluded_pixels=excluded_pixels)
    adjusted = corr.adjust_tileshape(
        tile_shape=tile_shape, sig_shape=sig_shape, base_shape=base_shape
    )
    assert adjusted == (123, 2)
    _validate(excluded_coords=excluded_coords, adjusted=adjusted, sig_shape=sig_shape)


@pytest.mark.with_numba
def test_tileshape_adjustment_6_3():
    sig_shape = (123, 456)
    tile_shape = (1, 1)
    base_shape = (1, 1)
    excluded_coords = np.array([
        range(123),
        range(0, 246, 2)
    ])
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=sig_shape, data=True)
    corr = CorrectionSet(excluded_pixels=excluded_pixels)
    adjusted = corr.adjust_tileshape(
        tile_shape=tile_shape, sig_shape=sig_shape, base_shape=base_shape
    )
    assert adjusted == (123, 246)
    _validate(excluded_coords=excluded_coords, adjusted=adjusted, sig_shape=sig_shape)


def test_tileshape_adjustment_7():
    sig_shape = (123, 456)
    tile_shape = (14, 42)
    base_shape = (7, 1)
    excluded_coords = np.array([
        (14, ),
        (42, )
    ])
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=sig_shape, data=True)
    corr = CorrectionSet(excluded_pixels=excluded_pixels)
    adjusted = corr.adjust_tileshape(
        tile_shape=tile_shape, sig_shape=sig_shape, base_shape=base_shape
    )
    assert adjusted == (21, 41)
    _validate(excluded_coords=excluded_coords, adjusted=adjusted, sig_shape=sig_shape)


def test_tileshape_adjustment_8():
    sig_shape = (1014, 1024)
    tile_shape = (1, 1)
    base_shape = (1, 1)
    # These magic numbers are "worst case" to produce collisions
    # 2*3*4*5*6*7
    excluded_coords = np.array([
        (720, 210, 306),
        (120, 210, 210)
    ])
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=sig_shape, data=True)
    corr = CorrectionSet(excluded_pixels=excluded_pixels)
    adjusted = corr.adjust_tileshape(
        tile_shape=tile_shape, sig_shape=sig_shape, base_shape=base_shape
    )
    print(adjusted)
    assert adjusted != (1014, 1024)
    _validate(excluded_coords=excluded_coords, adjusted=adjusted, sig_shape=sig_shape)


def test_tileshape_adjustment_9():
    sig_shape = (123, 456)
    tile_shape = (8, 1)
    base_shape = (2, 1)
    excluded_coords = np.array([
        (122, ),
        (455, )
    ])
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=sig_shape, data=True)
    corr = CorrectionSet(excluded_pixels=excluded_pixels)
    adjusted = corr.adjust_tileshape(
        tile_shape=tile_shape, sig_shape=sig_shape, base_shape=base_shape
    )
    assert adjusted == (8, 2)
    _validate(excluded_coords=excluded_coords, adjusted=adjusted, sig_shape=sig_shape)


def test_tileshape_adjustment_10():
    sig_shape = (122, 455)
    tile_shape = (8, 1)
    base_shape = (2, 1)
    excluded_coords = np.array([
        (121, ),
        (454, )
    ])
    excluded_pixels = sparse.COO(coords=excluded_coords, shape=sig_shape, data=True)
    corr = CorrectionSet(excluded_pixels=excluded_pixels)
    adjusted = corr.adjust_tileshape(
        tile_shape=tile_shape, sig_shape=sig_shape, base_shape=base_shape
    )
    assert adjusted == (8, 3)
    _validate(excluded_coords=excluded_coords, adjusted=adjusted, sig_shape=sig_shape)


def test_tileshape_adjustment_fuzz():
    for n in range(10):
        sig_shape = (np.random.randint(1, 2**12), np.random.randint(1, 2**12))
        print("Sig shape", sig_shape)
        tile_shape = (1, 1)
        base_shape = (1, 1)
        size = max(1, max(sig_shape) // 10)
        excluded_coords = np.vstack([
            np.random.randint(0, sig_shape[0], size=size),
            np.random.randint(0, sig_shape[1], size=size),
        ])
        print("excluded_coords", excluded_coords.shape, excluded_coords)
        excluded_pixels = sparse.COO(coords=excluded_coords, shape=sig_shape, data=True)
        corr = CorrectionSet(excluded_pixels=excluded_pixels, allow_empty=True)
        adjusted = corr.adjust_tileshape(
            tile_shape=tile_shape, sig_shape=sig_shape, base_shape=base_shape
        )
        print(adjusted)
        _validate(excluded_coords=excluded_coords, adjusted=adjusted, sig_shape=sig_shape)


class EarlyExit(Exception):
    pass


class EarlyExitUDF(NoOpUDF):
    def process_tile(self, tile):
        raise EarlyExit


def test_tileshape_adjustment_many(large_raw, lt_ctx):
    udf = EarlyExitUDF()
    exclude = sparse.COO(
        coords=exclude_pixels(sig_dims=tuple(large_raw.shape.sig), num_excluded=1000),
        shape=tuple(large_raw.shape.sig),
        data=True
    )
    corr = CorrectionSet(
        excluded_pixels=exclude
    )
    with pytest.raises(EarlyExit):
        lt_ctx.run_udf(dataset=large_raw, udf=udf, corrections=corr)



import typing
import os

import scipy.sparse
import numpy as np
import numba
import tomli
from sparseconverter import SCIPY_CSR, ArrayBackend, for_backend, NUMPY

from libertem.common import Slice, Shape
from libertem.common.math import prod, count_nonzero
from libertem.io.corrections.corrset import CorrectionSet
from libertem.io.dataset.base import (
    DataTile, DataSet
)
from libertem.io.dataset.base.meta import DataSetMeta
from libertem.io.dataset.base.partition import Partition
from libertem.io.dataset.base.tiling_scheme import TilingScheme
from libertem.common.messageconverter import MessageConverter
from libertem.common.numba import numba_dtypes

if typing.TYPE_CHECKING:
    from libertem.io.dataset.base.backend import IOBackend
    from libertem.common.executor import JobExecutor
    import numpy.typing as nt


def load_toml(path: str):
    with open(path, "rb") as f:
        return tomli.load(f)


class RawCSRDatasetParams(MessageConverter):
    SCHEMA = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "http://libertem.org/RawCSRDatasetParams.schema.json",
        "title": "RawCSRDatasetParams",
        "type": "object",
        "properties": {
            "type": {"const": "RAW_CSR"},
            "path": {"type": "string"},
            "nav_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sig_shape": {
                "type": "array",
                "items": {"type": "number", "minimum": 1},
                "minItems": 2,
                "maxItems": 2
            },
            "sync_offset": {"type": "number"},
        },
        "required": ["type", "path"]
    }

    def convert_to_python(self, raw_data):
        data = {
            k: raw_data[k]
            for k in ["path"]
        }
        if "nav_shape" in raw_data:
            data["nav_shape"] = tuple(raw_data["nav_shape"])
        if "sig_shape" in raw_data:
            data["sig_shape"] = tuple(raw_data["sig_shape"])
        if "sync_offset" in raw_data:
            data["sync_offset"] = raw_data["sync_offset"]
        return data


class CSRDescriptor(typing.NamedTuple):
    indptr_file: str
    indptr_dtype: np.dtype
    indices_file: str
    indices_dtype: np.dtype
    data_file: str
    data_dtype: np.dtype


class CSRTriple(typing.NamedTuple):
    indptr: np.ndarray
    indices: np.ndarray
    data: np.ndarray


class RawCSRDataSet(DataSet):
    """
    Read sparse data in compressed sparse row (CSR) format from a triple of files
    that contain the index pointers, the coordinates and the values. See
    `Wikipedia article on the CSR format
    <https://en.wikipedia.org/wiki/Sparse_matrix#Compressed_sparse_row_(CSR,_CRS_or_Yale_format)>`_
    for more information on the format.

    The necessary parameters are specified in a TOML file like this:

    .. code-block::

        [params]

        filetype = "raw_csr"
        nav_shape = [512, 512]
        sig_shape = [516, 516]

        [raw_csr]

        indptr_file = "rowind.dat"
        indptr_dtype = "<i4"

        indices_file = "coords.dat"
        indices_dtype = "<i4"

        data_file = "values.dat"
        data_dtype = "<i4"`

    Both the navigation and signal axis are flattened in the file, so that existing
    CSR libraries like scipy.sparse can be used directly by memory-mapping or
    reading the file contents.

    Parameters
    ----------

    path : str
        Path to the TOML file with file names and other parameters for the sparse dataset.
    nav_shape : Tuple[int, int], optional
        A nav_shape to apply to the dataset overriding the shape
        value read from the TOML file, by default None. This can
        be used to read a subset of the data, or reshape the
        contained data.
    sig_shape : Tuple[int, int], optional
        A sig_shape to apply to the dataset overriding the shape
        value read from the TOML file, by default None.
    sync_offset : int, optional, by default 0
        If positive, number of frames to skip from start
        If negative, number of blank frames to insert at start
    io_backend : IOBackend, optional
        The I/O backend to use, see :ref:`io backends`, by default None.

    Examples
    --------

    >>> ds = ctx.load("raw_csr", path='./path_to.toml')  # doctest: +SKIP
    """

    def __init__(
        self,
        path: str,
        nav_shape: typing.Optional[tuple[int, ...]] = None,
        sig_shape: typing.Optional[tuple[int, ...]] = None,
        sync_offset: int = 0,
        io_backend: typing.Optional["IOBackend"] = None
    ):
        if io_backend is not None:
            raise NotImplementedError()
        super().__init__(io_backend=io_backend)
        self._path = path
        if nav_shape is not None:
            nav_shape = tuple(nav_shape)
        self._nav_shape = nav_shape
        if sig_shape is not None:
            sig_shape = tuple(sig_shape)
        self._sig_shape = sig_shape
        self._sync_offset = sync_offset
        self._conf = None
        self._descriptor = None

    def initialize(self, executor: "JobExecutor") -> "DataSet":
        self._conf = conf = executor.run_function(load_toml, self._path)
        assert conf is not None
        if conf['params']['filetype'].lower() != 'raw_csr':
            raise ValueError(f"Filetype is not CSR, found {conf['params']['filetype']}")
        nav_shape = tuple(conf['params']['nav_shape'])
        sig_shape = tuple(conf['params']['sig_shape'])
        if self._nav_shape is None:
            self._nav_shape = nav_shape
        if self._sig_shape is None:
            self._sig_shape = sig_shape
        else:
            if prod(self._sig_shape) != prod(sig_shape):
                raise ValueError(f"Sig size mismatch between {self._sig_shape} and {sig_shape}.")

        shape = Shape(self._nav_shape + self._sig_shape, sig_dims=len(self._sig_shape))
        self._descriptor = descriptor = executor.run_function(get_descriptor, self._path)
        executor.run_function(
            check,
            descriptor=descriptor,
            nav_shape=self._nav_shape,
            sig_shape=self._sig_shape
        )
        image_count = executor.run_function(get_nav_size, descriptor=descriptor)
        self._image_count = image_count
        self._nav_shape_product = int(prod(self._nav_shape))
        self._sync_offset_info = self.get_sync_offset_info()
        self._meta = DataSetMeta(
            shape=shape,
            array_backends=[SCIPY_CSR],
            image_count=image_count,
            raw_dtype=descriptor.data_dtype,
            dtype=None,
            metadata=None,
            sync_offset=self._sync_offset,
        )
        return self

    @property
    def dtype(self) -> "nt.DTypeLike":
        assert self._meta is not None
        return self._meta.raw_dtype

    @property
    def shape(self) -> Shape:
        assert self._meta is not None
        return self._meta.shape

    @property
    def array_backends(self) -> typing.Sequence[ArrayBackend]:
        assert self._meta is not None
        return self._meta.array_backends

    def get_base_shape(self, roi):
        return (1, ) + tuple(self.shape.sig)

    def get_max_io_size(self):
        # High value since referring to dense for the time being
        # Compromise between memory use during densification and
        # performance with native sparse
        return int(1024*1024*20)

    def check_valid(self) -> bool:
        return True  # TODO

    @staticmethod
    def _get_filesize(path):
        return os.stat(path).st_size

    def supports_correction(self):
        return False

    @classmethod
    def detect_params(cls, path: str, executor: "JobExecutor"):
        try:
            _, extension = os.path.splitext(path)
            has_extension = extension.lstrip('.') in cls.get_supported_extensions()
            under_size_lim = executor.run_function(cls._get_filesize, path) < 2**20  # 1 MB
            if not (has_extension or under_size_lim):
                return False
            conf = executor.run_function(load_toml, path)
            if "params" not in conf:
                return False

            if "filetype" not in conf["params"]:
                return False
            if conf["params"]["filetype"].lower() != "raw_csr":
                return False
            descriptor = executor.run_function(get_descriptor, path)
            image_count = executor.run_function(get_nav_size, descriptor=descriptor)
            return {
                "parameters": {
                    'path': path,
                    "nav_shape": conf["params"]["nav_shape"],
                    "sig_shape": conf["params"]["sig_shape"],
                    "sync_offset": 0,
                },
                "info": {
                    "image_count": image_count,
                }
            }
        except (TypeError, UnicodeDecodeError, tomli.TOMLDecodeError, OSError):
            return False

    @classmethod
    def get_msg_converter(cls) -> type["MessageConverter"]:
        return RawCSRDatasetParams

    def get_diagnostics(self):
        return [
            {"name": "data dtype", "value": str(self._descriptor.data_dtype)},
            {"name": "indptr dtype", "value": str(self._descriptor.indptr_dtype)},
            {"name": "indices dtype", "value": str(self._descriptor.indices_dtype)},
        ]  # TODO: nonzero elements?

    @classmethod
    def get_supported_extensions(cls) -> set[str]:
        return {"toml"}

    def get_cache_key(self) -> str:
        raise NotImplementedError()  # TODO

    @classmethod
    def get_supported_io_backends(cls) -> list[str]:
        return []  # FIXME: we may want to read using a backend in the future

    def adjust_tileshape(
        self,
        tileshape: tuple[int, ...],
        roi: typing.Optional[np.ndarray]
    ) -> tuple[int, ...]:
        return (tileshape[0],) + tuple(self._sig_shape)

    def need_decode(
        self,
        read_dtype: "nt.DTypeLike",
        roi: typing.Optional[np.ndarray],
        corrections: typing.Optional[CorrectionSet]
    ) -> bool:
        return super().need_decode(read_dtype, roi, corrections)

    def get_partitions(self) -> typing.Generator[Partition, None, None]:
        assert self._meta is not None
        for part_slice, start, stop in self.get_slices():
            yield RawCSRPartition(
                descriptor=self._descriptor,
                meta=self._meta,
                partition_slice=part_slice,
                start_frame=start,
                num_frames=stop - start,
                io_backend=None,
                decoder=None,
            )


class RawCSRPartition(Partition):
    def __init__(
        self,
        descriptor: CSRDescriptor,
        start_frame: int,
        num_frames: int,
        *args,
        **kwargs
    ):
        self._descriptor = descriptor
        self._start_frame = start_frame
        self._num_frames = num_frames
        self._corrections = CorrectionSet()
        self._worker_context = None
        super().__init__(*args, **kwargs)

    def set_corrections(self, corrections: typing.Optional[CorrectionSet]):
        if corrections is not None and corrections.have_corrections():
            raise NotImplementedError("corrections not implemented for raw CSR data set")

    def validate_tiling_scheme(self, tiling_scheme: TilingScheme):
        if len(tiling_scheme) != 1:
            raise ValueError("Cannot slice CSR data in sig dimensions")

    def get_locations(self):
        # Allow using any worker by default
        return None

    def get_tiles(
        self,
        tiling_scheme: TilingScheme,
        dest_dtype="float32",
        roi=None,
        array_backend: typing.Optional[ArrayBackend] = None
    ):
        assert array_backend == SCIPY_CSR or array_backend is None
        tiling_scheme = tiling_scheme.adjust_for_partition(self)
        self.validate_tiling_scheme(tiling_scheme)
        triple = get_triple(self._descriptor)
        if self._corrections is not None and self._corrections.have_corrections():
            raise NotImplementedError(
                "corrections are not yet supported for raw CSR"
            )
        if roi is None:
            yield from read_tiles_straight(
                triple, self.slice, self.meta.sync_offset, tiling_scheme, dest_dtype
            )
        else:
            yield from read_tiles_with_roi(
                triple, self.slice, self.meta.sync_offset, tiling_scheme, roi, dest_dtype
            )


def sliced_indptr(triple: CSRTriple, partition_slice: Slice, sync_offset: int):
    assert len(partition_slice.shape.nav) == 1
    skip = min(0, partition_slice.origin[0] + sync_offset)
    indptr_start = max(0, partition_slice.origin[0] + sync_offset)
    indptr_stop = max(0, partition_slice.origin[0] + partition_slice.shape.nav[0] + 1 + sync_offset)
    return skip, triple.indptr[indptr_start:indptr_stop]


def get_triple(descriptor: CSRDescriptor) -> CSRTriple:
    data: np.ndarray = np.memmap(
        descriptor.data_file,
        dtype=descriptor.data_dtype,
        mode='r'
    )
    indices: np.ndarray = np.memmap(
        descriptor.indices_file,
        dtype=descriptor.indices_dtype,
        mode='r'
    )
    indptr: np.ndarray = np.memmap(
        descriptor.indptr_file,
        dtype=descriptor.indptr_dtype,
        mode='r'
    )

    return CSRTriple(
        indptr=indptr,
        indices=indices,
        data=data,
    )


def check(descriptor: CSRDescriptor, nav_shape, sig_shape, debug=False):
    triple = get_triple(descriptor)
    if triple.indices.shape != triple.data.shape:
        raise RuntimeError('Shape mismatch between data and indices.')
    if debug:
        assert np.min(triple.indices) >= 0
        assert np.max(triple.indices) < prod(sig_shape)
        assert np.min(triple.indptr) >= 0
        assert np.max(triple.indptr) == len(triple.indices)


def get_descriptor(path: str) -> CSRDescriptor:
    """
    Get a CSRDescriptor from the path to a toml sidecar file
    """
    conf = load_toml(path)
    assert conf is not None
    if conf['params']['filetype'].lower() != 'raw_csr':
        raise ValueError(f"Filetype is not CSR, found {conf['params']['filetype']}")

    base_path = os.path.dirname(path)
    # make sure the key is not case sensitive to follow the convention of
    # the Context.load() function.
    csr_key = conf['params']['filetype']
    csr_conf = conf[csr_key]
    return CSRDescriptor(
        indptr_file=os.path.join(base_path, csr_conf['indptr_file']),
        indptr_dtype=csr_conf['indptr_dtype'],
        indices_file=os.path.join(base_path, csr_conf['indices_file']),
        indices_dtype=csr_conf['indices_dtype'],
        data_file=os.path.join(base_path, csr_conf['data_file']),
        data_dtype=csr_conf['data_dtype'],
    )


def get_nav_size(descriptor: CSRDescriptor) -> int:
    '''
    To run efficiently on a remote worker for dataset initialization
    '''
    indptr = np.memmap(
        descriptor.indptr_file,
        dtype=descriptor.indptr_dtype,
        mode='r',
    )
    return len(indptr) - 1


def read_tiles_straight(
    triple: CSRTriple,
    partition_slice: Slice,
    sync_offset: int,
    tiling_scheme: TilingScheme,
    dest_dtype: np.dtype,
):
    assert len(tiling_scheme) == 1

    skip, indptr = sliced_indptr(
        triple,
        partition_slice=partition_slice,
        sync_offset=sync_offset
    )

    sig_shape = tuple(partition_slice.shape.sig)
    sig_size = partition_slice.shape.sig.size
    sig_dims = len(sig_shape)

    # Technically, one could use the slicing implementation of csr_matrix here.
    # However, it is slower, presumably because it takes a copy
    # Furthermore it provides a template to use an actual I/O backend here
    # instead of memory mapping.
    for indptr_start in range(0, len(indptr) - 1, tiling_scheme.depth):
        tile_start = indptr_start - skip  # skip is a negative value or 0
        indptr_stop = min(indptr_start + tiling_scheme.depth, len(indptr) - 1)
        if indptr_stop - indptr_start <= 0:
            continue

        indptr_slice = indptr[indptr_start:indptr_stop + 1]

        start = indptr[indptr_start]
        stop = indptr[indptr_stop]
        data = triple.data[start:stop]
        if dest_dtype != data.dtype:
            data = data.astype(dest_dtype)
        indices = triple.indices[start:stop]

        indptr_slice = indptr_slice - indptr_slice[0]
        arr = scipy.sparse.csr_matrix(
            (data, indices, indptr_slice),
            shape=(indptr_stop - indptr_start, sig_size)
        )
        tile_slice = Slice(
            origin=(partition_slice.origin[0] + tile_start, ) + (0, ) * sig_dims,
            shape=Shape((arr.shape[0], ) + sig_shape, sig_dims=sig_dims),
        )
        yield DataTile(
            data=arr,
            tile_slice=tile_slice,
            scheme_idx=0,
        )


def populate_tile(
    indptr_tile_start: "np.ndarray",
    indptr_tile_stop: "np.ndarray",
    orig_data: "np.ndarray",
    orig_indices: "np.ndarray",
    data_out: "np.ndarray",
    indices_out: "np.ndarray",
    indptr_out: "np.ndarray",
):
    offset = 0
    indptr_out[0] = 0
    for i, (start, stop) in enumerate(zip(indptr_tile_start, indptr_tile_stop)):
        chunk_size = stop - start
        data_out[offset:offset + chunk_size] = orig_data[start:stop]
        indices_out[offset:offset + chunk_size] = orig_indices[start:stop]
        offset += chunk_size
        indptr_out[i + 1] = offset


populate_tile_numba = numba.njit(populate_tile)


def can_use_numba(triple: CSRTriple) -> bool:
    return all(d in numba_dtypes
        for d in (triple.data.dtype, triple.indices.dtype, triple.indptr.dtype))


def read_tiles_with_roi(
    triple: CSRTriple,
    partition_slice: Slice,
    sync_offset: int,
    tiling_scheme: TilingScheme,
    roi: np.ndarray,
    dest_dtype: np.dtype,
):
    assert len(tiling_scheme) == 1
    roi = roi.reshape((-1, ))
    part_start = max(0, partition_slice.origin[0])
    tile_offset = count_nonzero(roi[:part_start])
    part_roi = partition_slice.get(roi, nav_only=True)

    skip, indptr = sliced_indptr(triple, partition_slice=partition_slice, sync_offset=sync_offset)

    if skip < 0:
        skipped_part_roi = part_roi[-skip:]
    else:
        skipped_part_roi = part_roi

    roi_overhang = max(0, len(skipped_part_roi) - len(indptr) + 1)
    if roi_overhang:
        real_part_roi = skipped_part_roi[:-roi_overhang]
    else:
        real_part_roi = skipped_part_roi

    real_part_roi = for_backend(real_part_roi, NUMPY)

    sig_shape = tuple(partition_slice.shape.sig)
    sig_size = partition_slice.shape.sig.size
    sig_dims = len(sig_shape)

    start_values = indptr[:-1][real_part_roi]
    stop_values = indptr[1:][real_part_roi]

    # Implementing this "by hand" instead of fancy indexing to provide a template to use an
    # actual I/O backend here instead of memory mapping.
    # The native scipy.sparse.csr_matrix implementation of fancy indexing
    # with a boolean mask for nav is very fast.

    if can_use_numba(triple):
        my_populate_tile = populate_tile_numba
    else:
        my_populate_tile = populate_tile

    for indptr_start in range(0, len(part_roi), tiling_scheme.depth):
        indptr_stop = min(indptr_start + tiling_scheme.depth, len(start_values))
        indptr_start = min(indptr_start, indptr_stop)
        # Don't read empty slices
        if indptr_stop - indptr_start <= 0:
            continue
        # Cast to int64 to avoid later upcasting to float64 in case of uint64
        # We can safely assume that files have less than 2**63 entries so that casting
        # from uint64 to int64 should be safe
        indptr_tile_start = start_values[indptr_start:indptr_stop].astype(np.int64)
        indptr_tile_stop = stop_values[indptr_start:indptr_stop].astype(np.int64)
        size = sum(indptr_tile_stop - indptr_tile_start)

        data = np.zeros(dtype=dest_dtype, shape=size)
        indices = np.zeros(dtype=triple.indices.dtype, shape=size)
        indptr_slice = np.zeros(
            dtype=indptr.dtype, shape=indptr_stop - indptr_start + 1
        )
        my_populate_tile(
            indptr_tile_start=indptr_tile_start,
            indptr_tile_stop=indptr_tile_stop,
            orig_data=triple.data,
            orig_indices=triple.indices,
            data_out=data,
            indices_out=indices,
            indptr_out=indptr_slice,
        )

        arr = scipy.sparse.csr_matrix(
            (data, indices, indptr_slice),
            shape=(indptr_stop - indptr_start, sig_size)
        )
        tile_slice = Slice(
            origin=(tile_offset + indptr_start, ) + (0, ) * sig_dims,
            shape=Shape((indptr_stop - indptr_start, ) + sig_shape, sig_dims=sig_dims),
        )
        yield DataTile(
            data=arr,
            tile_slice=tile_slice,
            scheme_idx=0,
        )



import os
import warnings

import defusedxml.ElementTree as ET
import numpy as np

from libertem.common.math import prod
from libertem.common import Shape
from libertem.common.messageconverter import MessageConverter
from .base import DataSet, DataSetException, DataSetMeta, BasePartition, IOBackend
from .raw import RawFile, RawFileSet

EMPAD_DETECTOR_SIZE = (128, 128)
EMPAD_DETECTOR_SIZE_RAW = (130, 128)


def get_params_from_xml(path, scan_parameters_mode="acquire"):
    """
    Parameters
    ----------

    path
        The path to the XML file (relative or absolute)

    scan_parameters_mode
        Either "acquire" or "search" - for some files, the raw
        data corresponds to the "search" scan parameters.
    """
    em = ET.parse(path)
    root = em.getroot()
    raw_filename = root.find("raw_file").attrib['filename']
    filename = os.path.basename(raw_filename)
    path_raw = os.path.join(
        os.path.dirname(path),
        filename
    )

    typ = root.find("type")

    # assume "scan" if no type is given:
    if typ is None or typ.text == 'scan':
        scan_parameters = [
            elem
            for elem in root.findall("scan_parameters")
            if elem.attrib["mode"] == scan_parameters_mode
        ]

        node_scan_x = scan_parameters[0].find("scan_resolution_x")
        node_scan_y = scan_parameters[0].find("scan_resolution_y")
        nav_x = int(node_scan_x.text)
        nav_y = int(node_scan_y.text)
        nav_shape = (nav_y, nav_x)
    elif typ.text == 'series':
        nav_shape = (int(root.find("count").text),)
    else:
        raise ValueError(f"unknown type: {typ.text}")

    return path_raw, nav_shape
    # TODO: read more metadata


class EMPADDatasetParams(MessageConverter):
    SCHEMA = {
      "$schema": "http://json-schema.org/draft-07/schema#",
      "$id": "http://libertem.org/EMPADDatasetParams.schema.json",
      "title": "EMPADDatasetParams",
      "type": "object",
      "properties": {
        "type": {"const": "EMPAD"},
        "path": {"type": "string"},
        "nav_shape": {
            "type": "array",
            "items": {"type": "number", "minimum": 1},
            "minItems": 2,
            "maxItems": 2
        },
        "sig_shape": {
            "type": "array",
            "items": {"type": "number", "minimum": 1},
            "minItems": 2,
            "maxItems": 2
        },
        "sync_offset": {"type": "number"},
        "io_backend": {
            "enum": IOBackend.get_supported(),
        },
      },
      "required": ["type", "path"]
    }

    def convert_to_python(self, raw_data):
        data = {
            k: raw_data[k]
            for k in ["path"]
        }
        if "nav_shape" in raw_data:
            data["nav_shape"] = tuple(raw_data["nav_shape"])
        if "sig_shape" in raw_data:
            data["sig_shape"] = tuple(raw_data["sig_shape"])
        if "sync_offset" in raw_data:
            data["sync_offset"] = raw_data["sync_offset"]
        return data


class EMPADFileSet(RawFileSet):
    def __init__(self, *args, **kwargs):
        kwargs.update({
            "frame_footer_bytes": 2*128*4,
        })
        super().__init__(*args, **kwargs)


class EMPADDataSet(DataSet):
    """
    Read data from EMPAD detector. EMPAD data sets consist of two files,
    one .raw and one .xml file. Note that the .xml file contains the file name
    of the .raw file, so if the raw file was renamed at some point, opening using
    the .xml file will fail.

    Parameters
    ----------
    path: str
        Path to either the .xml or the .raw file. If the .xml file given,
        the `nav_shape` parameter can be left out

    nav_shape: tuple of int, optional
        A tuple (y, x) or (num_images,) that specifies the size of the scanned
        region or number of frames in the series. It is automatically read from
        the .xml file if you specify one as `path`.

    sig_shape: tuple of int, optional
        Signal/detector size (height, width)

    sync_offset: int, optional
        If positive, number of frames to skip from start
        If negative, number of blank frames to insert at start

    Examples
    --------

    >>> ds = ctx.load("empad", path='./path_to_file.xml', ...)  # doctest: +SKIP
    """
    def __init__(self, path, scan_size=None, nav_shape=None,
                 sig_shape=None, sync_offset=0, io_backend=None):
        super().__init__(io_backend=io_backend)
        self._path = path
        self._nav_shape = tuple(nav_shape) if nav_shape else nav_shape
        self._sig_shape = tuple(sig_shape) if sig_shape else sig_shape
        self._sync_offset = sync_offset
        # handle backwards-compatability:
        if scan_size is not None:
            warnings.warn(
                "scan_size argument is deprecated. please specify nav_shape instead",
                FutureWarning
            )
            if nav_shape is not None:
                raise ValueError("cannot specify both scan_size and nav_shape")
            self._nav_shape = tuple(scan_size)
        self._path_raw = None
        self._meta = None

    def _init_from_xml(self, path, shape_fixup: bool = False):
        try:
            mode = "acquire"
            if shape_fixup:
                mode = "search"
            return get_params_from_xml(path, scan_parameters_mode=mode)
        except Exception as e:
            raise DataSetException(
                "could not initialize EMPAD file; error: %s" % (
                    str(e))
            )

    def initialize(self, executor):
        nav_shape_from_XML = None
        lowpath = self._path.lower()
        if lowpath.endswith(".xml"):
            self._path_raw, nav_shape_from_XML = executor.run_function(
                self._init_from_xml, self._path
            )
        else:
            if not lowpath.endswith(".raw"):
                raise DataSetException("path should either be .xml or .raw")
            if self._nav_shape is None:
                raise DataSetException("need to set or detect nav_shape!")
            self._path_raw = self._path

        try:
            self._filesize = executor.run_function(self._get_filesize)
        except OSError as e:
            raise DataSetException(f"could not open file {self._path_raw}: {str(e)}")

        itemsize = int(np.dtype("float32").itemsize)

        # consistency check: see that the file size matches the detected nav
        # shape, otherwise, re-run with fixup for #1617
        if nav_shape_from_XML is not None:
            expected_size = (
                prod(EMPAD_DETECTOR_SIZE_RAW) * prod(nav_shape_from_XML) * itemsize
            )
            alternate_size = None
            alternate_nav_shape_from_XML = None
            if self._filesize != expected_size:
                # Extract alternate nav shape with shape_fixup=True
                _, alternate_nav_shape_from_XML = executor.run_function(
                    self._init_from_xml, self._path, shape_fixup=True,
                )
                alternate_size = (
                    prod(EMPAD_DETECTOR_SIZE_RAW) * prod(alternate_nav_shape_from_XML) * itemsize
                )
            if self._filesize != expected_size and self._filesize != alternate_size:
                raise ValueError(
                    f"RAW data file size mismatch; filesize={self._filesize} "
                    f"vs expected size {expected_size} for nav {nav_shape_from_XML} "
                    f"or alternate {alternate_size} for nav {alternate_nav_shape_from_XML}."
                )
            elif self._filesize == alternate_size:
                # apply mitigation
                nav_shape_from_XML = alternate_nav_shape_from_XML

        self._image_count = int(
            self._filesize / (
                itemsize * int(
                    prod(EMPAD_DETECTOR_SIZE_RAW)
                )
            )
        )
        if self._nav_shape is None and nav_shape_from_XML is not None:
            self._nav_shape = nav_shape_from_XML
        elif self._nav_shape is None and nav_shape_from_XML is None:
            raise ValueError(
                "either nav_shape needs to be passed, or path needs to point to the .xml file"
            )
        self._nav_shape_product = int(prod(self._nav_shape))
        if nav_shape_from_XML:
            self._image_count = int(prod(nav_shape_from_XML))
        if self._sig_shape is None:
            self._sig_shape = EMPAD_DETECTOR_SIZE
        elif int(prod(self._sig_shape)) != int(prod(EMPAD_DETECTOR_SIZE)):
            raise DataSetException(
                "sig_shape must be of size: %s" % int(prod(EMPAD_DETECTOR_SIZE))
            )
        self._sync_offset_info = self.get_sync_offset_info()
        self._meta = DataSetMeta(
            shape=Shape(self._nav_shape + self._sig_shape, sig_dims=len(self._sig_shape)),
            raw_dtype=np.dtype("float32"),
            sync_offset=self._sync_offset,
            image_count=self._image_count,
        )
        return self

    def _get_filesize(self):
        return os.stat(self._path_raw).st_size

    @classmethod
    def get_msg_converter(cls):
        return EMPADDatasetParams

    @classmethod
    def get_supported_extensions(cls):
        return {"xml", "raw"}

    @classmethod
    def detect_params(cls, path, executor):
        """
        Detect parameters. If an `path` is an xml file, we try to automatically
        set the nav_shape, otherwise we can't really detect if this is a EMPAD
        file or something else (maybe from the "trailer" after each frame?)
        """
        try:
            ds = cls(path)
            ds = ds.initialize(executor)
            if not executor.run_function(ds.check_valid):
                return False
            return {
                "parameters": {
                    "path": path,
                    "nav_shape": ds._nav_shape,
                    "sig_shape": ds._sig_shape,
                },
                "info": {
                    "image_count": ds._image_count,
                    "native_sig_shape": ds._sig_shape,
                }
            }
        except Exception:
            return False

    @property
    def dtype(self):
        return self._meta.raw_dtype

    @property
    def shape(self):
        return self._meta.shape

    def _get_fileset(self):
        return EMPADFileSet([
            RawFile(
                path=self._path_raw,
                start_idx=0,
                end_idx=self._image_count,
                sig_shape=self.shape.sig,
                native_dtype=self._meta.raw_dtype,
                frame_footer=2*128*4,
            )
        ])

    def check_valid(self):
        try:
            fileset = self._get_fileset()
            backend = self.get_io_backend().get_impl()
            with backend.open_files(fileset):
                return True
        except (OSError, ValueError) as e:
            raise DataSetException("invalid dataset: %s" % e)

    def get_cache_key(self):
        return {
            "path_raw": self._path_raw,
            "shape": tuple(self.shape),
            "sync_offset": self._sync_offset,
        }

    def get_partitions(self):
        fileset = self._get_fileset()
        for part_slice, start, stop in self.get_slices():
            yield BasePartition(
                meta=self._meta,
                fileset=fileset,
                partition_slice=part_slice,
                start_frame=start,
                num_frames=stop - start,
                io_backend=self.get_io_backend(),
            )

    def __repr__(self):
        return f"<EMPADFileDataSet of {self.dtype} shape={self.shape}>"



from libertem.io.dataset import detect
from libertem.executor.inline import InlineJobExecutor


def test_detection_empty_hdf5(empty_hdf5):
    executor = InlineJobExecutor()
    fn = empty_hdf5.filename
    params = detect(fn, executor=executor)["parameters"]
    assert params != {}
    assert list(params.keys()) == ["path"]


def test_detection_nonempty_hdf5(hdf5_ds_1):
    executor = InlineJobExecutor()
    fn = hdf5_ds_1.path
    params = detect(fn, executor=executor)
    parameters = params["parameters"]
    assert parameters != {}
    assert parameters["ds_path"] == "data"
    assert parameters["path"] == fn
    assert params["type"] == "hdf5"
    assert list(parameters.keys()) == ["path", "ds_path", "nav_shape", "sig_shape"]



import numpy as np
import inspect

from libertem import masks
from .masks import SingleMaskAnalysis
from .helper import GeneratorHelper


class DiskTemplate(GeneratorHelper):

    short_name = "disk"
    api = "create_disk_analysis"

    def __init__(self, params):
        self.params = params

    def get_docs(self):
        title = "Disk Analysis"
        from libertem.api import Context
        docs_rst = inspect.getdoc(Context.create_disk_analysis)
        docs = self.format_docs(title, docs_rst)
        return docs

    def convert_params(self):
        params = ['dataset=ds']
        for k in ['cx', 'cy', 'r']:
            params.append(f'{k}={self.params[k]}')
        return ', '.join(params)

    def get_plot(self):
        plot = ["plt.figure()",
                "plt.imshow(disk_result['intensity'])",
                "plt.colorbar()"]
        return ['\n'.join(plot)]


class DiskMaskAnalysis(SingleMaskAnalysis, id_="APPLY_DISK_MASK"):
    TYPE = 'UDF'

    def get_description(self):
        return "intensity of the integration over the selected disk"

    def get_mask_factories(self):
        if self.dataset.shape.sig.dims != 2:
            raise ValueError("can only handle 2D signals currently")
        (detector_y, detector_x) = self.dataset.shape.sig

        cx = self.parameters['cx']
        cy = self.parameters['cy']
        r = self.parameters['r']

        def disk_mask():
            return masks.circular(
                centerX=cx,
                centerY=cy,
                imageSizeX=detector_x,
                imageSizeY=detector_y,
                radius=r,
            )

        return [
            disk_mask,
        ]

    def get_parameters(self, parameters):
        (detector_y, detector_x) = self.dataset.shape.sig

        cx = parameters.get('cx', detector_x / 2)
        cy = parameters.get('cy', detector_y / 2)
        r = parameters.get('r', min(detector_y, detector_x) / 2 * 0.3)
        use_sparse = parameters.get('use_sparse', False)

        return {
            'cx': cx,
            'cy': cy,
            'r': r,
            'use_sparse': use_sparse,
            'mask_count': 1,
            'mask_dtype': np.float32,
        }

    def need_rerun(self, old_params, new_params):
        return old_params != new_params

    @classmethod
    def get_template_helper(cls):
        return DiskTemplate



import numpy as np
import sparse
import inspect
import re
from textwrap import dedent

from .base import BaseAnalysis, AnalysisResult, AnalysisResultSet
from libertem.common.async_utils import sync_to_async
from libertem.common.executor import JobCancelledError
from libertem.udf.masks import ApplyMasksUDF
from libertem.udf.stddev import StdDevUDF, consolidate_result
from .getroi import get_roi
from libertem.masks import _make_circular_mask
from .helper import GeneratorHelper


class ClusterTemplate(GeneratorHelper):

    short_name = "cluster"

    def __init__(self, params):
        self.params = params

    def format_code(self, code):
        code = re.sub(r"self(\.|, |)|(, |)executor|, cancel_id|, job_is_cancelled", "", code)
        # remove indentation
        code = dedent(code)
        return code

    def get_dependency(self):
        dep = [
           "from libertem.analysis import ClusterAnalysis",
           "from libertem.udf.stddev import StdDevUDF",
           "from libertem.masks import _make_circular_mask",
           "from skimage.feature import peak_local_max",
           "import sparse",
           "from sklearn.feature_extraction.image import grid_to_graph",
           "from libertem.udf.masks import ApplyMasksUDF",
           "from libertem.analysis.getroi import get_roi",
           "from sklearn.cluster import AgglomerativeClustering",
           "from libertem.analysis.base import AnalysisResult, AnalysisResultSet",
           "from libertem.viz import visualize_simple",
        ]
        return dep

    def temp_cluster_controller(self):
        temp_controller = [
                        "dataset = ds",
                        "roi, sd_udf_results = get_sd_results()",
                        "cluster_udf = get_cluster_udf(sd_udf_results)",
                        "udf_results = ctx.run_udf(dataset=ds, udf=cluster_udf, progress=True)",
                        "cluster_result = get_udf_results(udf_results, roi, damage=True)",
        ]
        return '\n'.join(temp_controller)

    def get_docs(self):
        docs = ["# Cluster Analysis"]
        return '\n'.join(docs)

    def get_run_sd_udf(self):
        indent = " "*4
        temp_run_sd = [
            "def run_sd_udf(roi, stddev_udf):",
            f"{indent}sd_udf_results = ctx.run_udf(dataset=ds, udf=stddev_udf, roi=roi,"
            " progress=True)",
            f"{indent}return roi, sd_udf_results"
        ]
        return "\n".join(temp_run_sd)

    def get_analysis(self):
        udf_results = self.format_code(inspect.getsource(ClusterAnalysis.get_udf_results))
        cluster_udf = self.format_code(inspect.getsource(ClusterAnalysis.get_cluster_udf))
        sd_results = self.format_code(inspect.getsource(ClusterAnalysis.get_sd_results))
        run_sd_udf = self.get_run_sd_udf()
        cluster_controller = self.temp_cluster_controller()

        temp_analysis = [
                    f"parameters={self.params}\ndataset=ds",
                    f"{run_sd_udf}",
                    f"{sd_results}",
                    f"{cluster_udf}",
                    f"{udf_results}",
                    f"{cluster_controller}"
                    ]
        return '\n\n'.join(temp_analysis)

    def get_plot(self):
        plot = ["plt.figure()",
                "plt.imshow(cluster_result['intensity'].raw_data)"]
        return ['\n'.join(plot)]


class ClusterAnalysis(BaseAnalysis, id_="CLUST"):
    TYPE = "UDF"

    def get_udf(self):
        # FIXME: we don't have all parameters available here to actually construct
        # a useful udf instance - this should be optional for Analysis subclasses
        # that implement the `controller` method
        return None

    def get_udf_results(self, udf_results, roi, damage):
        from libertem.viz import visualize_simple
        from sklearn.cluster import AgglomerativeClustering
        from sklearn.feature_extraction.image import grid_to_graph
        n_clust = self.parameters["n_clust"]
        y, x = tuple(self.dataset.shape.nav)
        connectivity = grid_to_graph(
            # Transposed!
            n_x=y,
            n_y=x
        )
        f = udf_results['intensity'].raw_data
        feature_vector = f / np.abs(f).mean(axis=0)

        clustering = AgglomerativeClustering(
            metric='euclidean', n_clusters=n_clust, linkage='ward',
            connectivity=connectivity
        ).fit(feature_vector)
        labels = np.array(clustering.labels_+1)
        return AnalysisResultSet([
            AnalysisResult(raw_data=labels.reshape(self.dataset.shape.nav),
                           visualized=visualize_simple(
                               labels.reshape(self.dataset.shape.nav),
                               damage=damage
                            ),
                           key="intensity", title="intensity",
                           desc="Labels for the clustering"),
        ])

    async def run_sd_udf(self, roi, stddev_udf, executor, cancel_id, job_is_cancelled):
        udf_runner_cls = executor.get_udf_runner()
        result_iter = udf_runner_cls([stddev_udf]).run_for_dataset_async(
            self.dataset, executor, roi=roi, cancel_id=cancel_id
        )
        async for sd_udf_results in result_iter:
            pass
        if job_is_cancelled():
            raise JobCancelledError()
        return roi, consolidate_result(sd_udf_results.buffers[0])

    def get_sd_results(self, executor, cancel_id, job_is_cancelled):
        stddev_udf = StdDevUDF()
        roi = get_roi(params=self.parameters, shape=self.dataset.shape.nav)
        return self.run_sd_udf(roi, stddev_udf, executor, cancel_id, job_is_cancelled)

    def get_cluster_udf(self, sd_udf_results):
        from skimage.feature import peak_local_max
        center = (self.parameters["cy"], self.parameters["cx"])
        rad_in = self.parameters["ri"]
        rad_out = self.parameters["ro"]
        n_peaks = self.parameters["n_peaks"]
        min_dist = self.parameters["min_dist"]
        sstd = sd_udf_results['std'].data
        sshape = sstd.shape
        if not (center is None or rad_in is None or rad_out is None):
            mask_out = 1*_make_circular_mask(center[1], center[0], sshape[1], sshape[0], rad_out)
            mask_in = 1*_make_circular_mask(center[1], center[0], sshape[1], sshape[0], rad_in)
            mask = mask_out - mask_in
            masked_sstd = sstd*mask
        else:
            masked_sstd = sstd

        coordinates = peak_local_max(masked_sstd, num_peaks=n_peaks, min_distance=min_dist)

        y = coordinates[..., 0]
        x = coordinates[..., 1]
        z = range(len(y))

        mask = sparse.COO(
            shape=(len(y), ) + tuple(self.dataset.shape.sig),
            coords=(z, y, x), data=1.
        )

        udf = ApplyMasksUDF(
            # float32 for cupy support
            mask_factories=lambda: mask, mask_count=len(y), mask_dtype=np.float32,
            use_sparse=True
        )
        return udf

    async def controller(self, cancel_id, executor, job_is_cancelled, send_results):
        roi, sd_udf_results = await self.get_sd_results(executor, cancel_id, job_is_cancelled)
        udf = self.get_cluster_udf(sd_udf_results)

        udf_runner_cls = executor.get_udf_runner()
        result_iter = udf_runner_cls([udf]).run_for_dataset_async(
            self.dataset, executor, cancel_id=cancel_id
        )
        async for udf_results in result_iter:
            pass

        if job_is_cancelled():
            raise JobCancelledError()

        results = await sync_to_async(
            self.get_udf_results,
            udf_results=udf_results.buffers[0],
            roi=roi,
            damage=udf_results.damage
        )
        await send_results(results, True)

    @classmethod
    def get_template_helper(cls):
        return ClusterTemplate



import itertools
import os
import sys
import logging
import functools
import contextlib
import multiprocessing as mp
from typing import (
    TYPE_CHECKING, Any, Callable, NamedTuple, Optional, TypeVar, Union,
)
from collections.abc import Iterable, Generator
from typing_extensions import TypedDict, Literal
import uuid
import warnings
import time

from tblib import pickling_support
import cloudpickle
from opentelemetry import trace
from libertem.common.backend import set_use_cpu, set_use_cuda, set_file_limit

from libertem.common.executor import (
    Environment, TaskProtocol, WorkerContext, WorkerQueue,
    WorkerQueueEmpty, TaskCommHandler, SimpleMPWorkerQueue,
    JobCancelledError, ResourceDef,
)
from libertem.common.scheduler import Worker, WorkerSet, Scheduler
from libertem.common.tracing import add_partition_to_span, attach_to_parent, maybe_setup_tracing

from .utils import assign_cudas
from .base import BaseJobExecutor, ResourceError

try:
    import prctl
except ImportError:
    prctl = None

if TYPE_CHECKING:
    from opentelemetry.trace import SpanContext

tracer = trace.get_tracer(__name__)
logger = logging.getLogger(__name__)


T = TypeVar('T')


class WorkerQueues(NamedTuple):
    request: "WorkerQueue"
    response: "WorkerQueue"
    message: "WorkerQueue"


class WorkerSpec(TypedDict):
    name: str
    device_id: int
    device_kind: Union[Literal['CPU'], Literal['CUDA']]
    worker_idx: int
    has_cupy: bool


class PoolWorkerInfo(NamedTuple):
    queues: WorkerQueues
    process: mp.Process
    spec: WorkerSpec


class PoolStateError(Exception):
    """
    The worker pool is not in the expected state to perform the requested operation
    """
    pass


def _resources_for_spec(spec: WorkerSpec) -> ResourceDef:
    worker_resources: "ResourceDef" = {}

    if spec['device_kind'] == 'CPU':
        worker_resources["compute"] = 1
        worker_resources["CPU"] = 1
        worker_resources["ndarray"] = 1

    if spec["device_kind"] == "CUDA":
        worker_resources["compute"] = 1
        worker_resources["CUDA"] = 1
        if spec["has_cupy"]:
            worker_resources["ndarray"] = 1

    return worker_resources


class WorkerPool:
    """
    Combination of worker processes and matching request queues,
    and a single response queue.

    Processes are started using the spawn method, so they need
    to use primitives that are compatible with spawn.

    Take care to properly close queues and join workers at shutdown.

    Note
    ----
    We are not using the vanilla :class:`multiprocesing.Pool` here, because
    we need to coordinate the sending and receiving side for each task,
    and we need to keep state on the worker, pin processes to cores etc.
    """
    def __init__(self, worker_fn: Callable, spec: list[WorkerSpec]):
        self._worker_q_cls = SimpleMPWorkerQueue
        self._workers: Optional[list[PoolWorkerInfo]] = []
        self._worker_fn = worker_fn
        self._response_q: Optional[SimpleMPWorkerQueue] = self._worker_q_cls()
        self._message_q = self._worker_q_cls()
        self._mp_ctx = mp.get_context("spawn")
        self._spec = spec
        self._start_workers()

    @property
    def response_queue(self) -> "WorkerQueue":
        if self._response_q is None:
            raise PoolStateError("Response queue not available")
        return self._response_q

    @property
    def message_queue(self) -> "WorkerQueue":
        return self._message_q

    @property
    def size(self) -> int:
        return len(self._spec)

    def _start_workers(self):
        with tracer.start_as_current_span("WorkerPool._start_workers") as span:
            span_context = span.get_span_context()
            for spec_item in self._spec:
                queues = self._make_worker_queues()
                p = self._mp_ctx.Process(target=self._worker_fn, kwargs={
                    "queues": queues,
                    "spec": spec_item,
                    "span_context": span_context,
                })
                p.start()
                self._workers.append(
                    PoolWorkerInfo(
                        queues=queues,
                        process=p,
                        spec=spec_item,
                    )
                )

    def kill_worker(self, worker_info: PoolWorkerInfo, timeout: float = 5.0):
        worker_info.queues.request.close(drain=False, force=True)
        worker_info.process.terminate()
        worker_info.process.join(timeout)
        if worker_info.process.exitcode is None:
            worker_info.process.kill()
            # reap the dead process:
            worker_info.process.join(30)

    def kill(self, timeout: float = 5):
        if self._workers is None:
            raise PoolStateError("Cannot kill workers, not running")
        for worker in self._workers:
            self.kill_worker(worker, timeout=timeout)
        exitcodes = [
            worker.process.exitcode
            for worker in self._workers
        ]
        self._workers = None
        if self._response_q is None:
            raise PoolStateError("Response queue not available")
        self._response_q.close(drain=False, force=True)
        self._response_q = None
        if any([e is None for e in exitcodes]):
            raise PoolStateError("At least one worker failed to be killed")

    @property
    def workers(self) -> list[PoolWorkerInfo]:
        if self._workers is None:
            raise PoolStateError("No workers are running")
        return self._workers

    def all_alive(self) -> bool:
        if self._workers is None:
            raise PoolStateError("No workers are running")
        return all(qp.process.is_alive() for qp in self._workers)

    def assert_all_alive(self):
        """
        Check if any workers are dead, if so, kill the whole pool and raise a
        `RuntimeError`.
        """
        if not self.all_alive():
            self.kill()
            raise RuntimeError(
                "One or more workers failed to start"
            )

    def close_resp_queue(self):
        self._response_q.close()

    def close_mesg_queue(self):
        self._message_q.close()

    def get_worker_queues(self, worker_idx: int) -> WorkerQueues:
        if self._workers is None:
            raise PoolStateError("No workers are running")
        return self._workers[worker_idx].queues

    def _make_worker_queues(self):
        return WorkerQueues(
            request=self._worker_q_cls(),
            response=self._response_q,
            message=self._message_q,
        )


WorkerSelector = Callable[
    [list[PoolWorkerInfo], WorkerPool, TaskProtocol, int],
    PoolWorkerInfo
]


def _select_by_queue_size(
    eligible: list[PoolWorkerInfo],
    pool: WorkerPool,
    task: TaskProtocol,
    task_idx: int,
) -> PoolWorkerInfo:
    # FIXME: I think this introduces some latency currently,
    # as the workers `get` from their queue as soon as possible,
    # so the following could happen:
    # - task 0 becomes available
    # - all queues have size 0, so task gets assigned to worker 0
    # - worker 0 picks the task from its queue, so its queue size is 0 again
    # - task 1 becomes available
    # - all queues have size 0 (again), so task gets assigned to worker 0
    # => this should only happen in the offline case, as in the live
    #    processing case, we push more than one message per task into the
    #    queue, so there should be more than one message in the backlog
    #    when the `TaskCommHandler` is finished pushing messages for one task.
    return min(eligible, key=lambda w: w.queues.request.size())


def _select_by_round_robin(
    eligible: list[PoolWorkerInfo],
    pool: WorkerPool,
    task: TaskProtocol,
    task_idx: int,
) -> PoolWorkerInfo:
    return eligible[task_idx % len(eligible)]


def schedule_task(
    task_idx: int,
    task: TaskProtocol,
    pool: WorkerPool,
    scheduler: Scheduler,
    pool_info_for_worker: dict[Worker, PoolWorkerInfo],
    selector: WorkerSelector,
) -> tuple[int, WorkerQueues]:
    """
    Returns the worker index and its queues of the worker the given task should
    be scheduled on.

    Workers are first filtered based on requested resources, then the concrete
    worker is selected by the function specified in the :code:`selector`
    parameter.
    """

    eligible = [
        pool_info_for_worker[w]
        for w in scheduler.workers_for_task(task)
    ]
    if len(eligible) == 0:
        resources = task.get_resources()
        raise ResourceError(f"No worker available with resources {resources}")

    worker = selector(eligible, pool, task, task_idx)

    idx = pool.workers.index(worker)
    return idx, worker.queues


def set_thread_name(name: str):
    """
    Set a thread name; mostly useful for using system tools for profiling

    Parameters
    ----------
    name : str
        The thread name
    """
    if prctl is None:
        return
    prctl.set_name(name)


def worker_run_task(
    header: dict,
    work_mem: dict,
    queues: WorkerQueues,
    worker_idx: int,
    env: Environment,
):
    """
    Called from the worker main loop when a RUN_TASK message is received

    Parameters
    ----------

    header
        The header of the message that was received

    work_mem
        The worker's working memory, for accessing scattered data

    queues
        request and response queues

    worker_idx
        This worker's index

    env
        The Environment for preparing thread counts etc. for the UDF run
    """
    with tracer.start_as_current_span("RUN_TASK") as span:
        try:
            span.set_attributes({
                "libertem.task_size_pickled": len(header["task"]),
                "libertem.os.pid": os.getpid(),
            })
            task: TaskProtocol = cloudpickle.loads(header["task"])
            params_handle = header["params_handle"]
            params = work_mem[params_handle]
            partition = task.get_partition()
            add_partition_to_span(partition)
            result = task(params, env)
            queues.response.put({
                "type": "RESULT",
                "result": result,
                "task_id": header["task_id"],
                "uuid": header["uuid"],
                "worker_id": worker_idx,
            })
        except Exception as e:
            logger.exception("failure in RUN_TASK")
            pickling_support.install(e)
            queues.response.put({
                "type": "ERROR",
                "error": e,
                "exception": e,
                "worker_id": worker_idx,
                "uuid": header["uuid"],
            })
        finally:
            pass


def worker_run_function(header, queues, worker_idx):
    """
    Called from the worker main loop when a RUN_FUNCTION message is received

    Parameters
    ----------
    header
        The header of the message that was received

    queues
        request and response queues

    worker_idx
        This worker's index
    """
    with tracer.start_as_current_span("RUN_FUNCTION"):
        try:
            fn = cloudpickle.loads(header["fn"])
            result = fn()
            queues.response.put({
                "type": "RUN_FUNCTION_RESULT",
                "result": result,
                "worker_id": worker_idx,
            })
        except Exception as e:
            logger.exception("failure in RUN_FUNCTION")
            pickling_support.install(e)
            queues.response.put({
                "type": "ERROR",
                "error": e,
                "exception": e,
                "worker_id": worker_idx,
            })


def _drain_after_task(queues: WorkerQueues):
    """
    In case of an error, need to drain the request queue
    (anything that is left over from the detector-specific
    data that was sent in `TaskCommHandler.handle_task`):
    """
    with tracer.start_as_current_span("drain after task") as span:
        while True:
            with queues.request.get() as msg:
                header, payload = msg
                header_type = header["type"]
                span.add_event("msg", {"type": header_type})
                if header_type in (
                    "RUN_TASK", "SCATTER", "SCATTER_UPDATE",
                    "SCATTER_UPDATE_PATCH", "RUN_FUNCTION",
                    "DELETE", "SHUTDOWN", "WARMUP",
                ):
                    raise RuntimeError(
                        f"unexpected message type {header_type}"
                    )
                if header_type == "END_TASK":
                    break


def worker_loop(
    queues: WorkerQueues,
    work_mem: dict,
    worker_idx: int,
    env: Environment
):
    """
    The worker main loop, called when the worker setup is done.
    Waits for messages on the request queue until a SHUTDOWN message
    is received.

    Parameters
    ----------

    work_mem
        The worker's working memory, for accessing scattered data

    queues
        request and response queues

    worker_idx
        This worker's index

    env
        The Environment for preparing thread counts etc. for the UDF run
    """
    while True:
        try:
            with queues.request.get() as msg:
                header, payload = msg
                header_type = header["type"]
                if header_type == "RUN_TASK":
                    with attach_to_parent(header["span_context"]):
                        worker_run_task(header, work_mem, queues, worker_idx, env)
                        _drain_after_task(queues)
                elif header_type == "SCATTER":
                    # FIXME: array data could be transferred and stored in SHM instead
                    key = header["key"]
                    if key in work_mem:
                        queues.response.put({
                            "type": "ERROR",
                            "error": f"key {key} already stored in worker memory",
                            "worker_id": worker_idx,
                        })
                        continue
                    work_mem[key] = header["value"]
                    continue
                elif header_type == "SCATTER_UPDATE":
                    # FIXME: array data could be transferred and stored in SHM instead
                    key = header["key"]
                    if key not in work_mem:
                        queues.response.put({
                            "type": "ERROR",
                            "error": f"key {key} not stored in worker memory, can't update",
                            "worker_id": worker_idx,
                        })
                        continue
                    work_mem[key] = header["value"]
                    continue
                elif header_type == "SCATTER_UPDATE_PATCH":
                    # FIXME: array data could be transferred and stored in SHM instead
                    key = header["key"]
                    if key not in work_mem:
                        queues.response.put({
                            "type": "ERROR",
                            "error": f"key {key} not stored in worker memory, can't update",
                            "worker_id": worker_idx,
                        })
                        continue
                    if not hasattr(work_mem[key], 'patch'):
                        queues.response.put({
                            "type": "ERROR",
                            "error": f"key {key} is not patcheable",
                            "worker_id": worker_idx,
                        })
                        continue
                    work_mem[key].patch(header["patch"])
                    continue
                elif header_type == "RUN_FUNCTION":
                    with attach_to_parent(header["span_context"]):
                        worker_run_function(header, queues, worker_idx)
                elif header_type == "DELETE":
                    key = header["key"]
                    if key in work_mem:
                        del work_mem[key]
                    continue
                elif header_type == "SHUTDOWN":
                    with attach_to_parent(header["span_context"]):
                        with tracer.start_as_current_span("SHUTDOWN"):
                            queues.request.close()
                            queues.response.close(drain=False)
                            queues.message.close(drain=False)
                        break
                elif header_type == "WARMUP":
                    with attach_to_parent(header["span_context"]):
                        with tracer.start_as_current_span("WARMUP"):
                            import libertem.udf.base  # NOQA
                            import libertem.api  # NOQA
                            import libertem.preload  # NOQA
                            with env.enter():
                                pass
                else:
                    queues.response.put({
                        "type": "ERROR",
                        "error": f"unknown message {header}",
                        "worker_id": worker_idx,
                    })
                    # probably desynchronized with the main process, so give up:
                    raise RuntimeError(f"unknown message, shutting down worker {worker_idx}")
        except KeyboardInterrupt:
            queues.response.put({
                "type": "ERROR",
                "error": "KeyboardInterrupt",
                "worker_id": worker_idx,
            })
            raise


DeviceT = tuple[int, Union[Literal['CUDA'], Literal['CPU']]]


def _setup_device(spec: WorkerSpec, pin: bool):
    """
    Set up this worker for its given task - either CPU or GPU comptation,
    and maybe pin CPU workers to a given CPU core.
    """
    if spec["device_kind"].lower() == "cpu":
        if hasattr(os, 'sched_setaffinity') and pin:
            os.sched_setaffinity(0, [spec["device_id"]])
        set_use_cpu(spec["device_id"])
    elif spec["device_kind"].lower() == "cuda":
        set_use_cuda(spec["device_id"])
    else:
        raise RuntimeError(f"unknown device kind: {spec['device_kind']}")


def pipelined_worker(
    queues: WorkerQueues,
    pin: bool,
    spec: WorkerSpec,
    span_context: "SpanContext",
    early_setup: Optional[Callable] = None,
):
    """
    Main pipelined worker function.

    Parameters
    ----------

    queues
        request and response queues

    pin
        Whether or not the CPU worker should be pinned to a specific CPU

    spec
        This worker's spec, containing name, worker index, device kind etc.

    span_context
        The tracing span we should attach to for the setup code

    early_setup
        Function that will be called very early in the setup code,
        allowing to inject custom functionality or specific warmup code
    """
    # FIXME: propagate to parent process with a pipe or similar?
    sys.stderr.close()
    sys.stderr = open(os.open(os.devnull, os.O_RDWR), closefd=False)

    try:
        if early_setup:
            early_setup()

        set_file_limit()

        # FIXME: explicitly propagate exporter settings to the workers?
        # right now taken care of by environment variables...
        worker_idx = spec["worker_idx"]
        maybe_setup_tracing(service_name="pipelined_worker", service_id=f"worker-{worker_idx}")

        # attach to parent span context for startup:
        with attach_to_parent(span_context), \
                tracer.start_as_current_span("pipelined_worker.startup") as span:
            _setup_device(spec, pin)
            work_mem: dict[str, Any] = {}

            span.set_attributes({
                "libertem.spec.name": spec["name"],
                "libertem.spec.device_id": spec["device_id"],
                "libertem.spec.devide_kind": spec["device_kind"],
                "libertem.spec.worker_idx": spec["worker_idx"],
            })

            set_thread_name(f"worker-{worker_idx}")

            worker_context = PipelinedWorkerContext(queues.request, queues.message)
            env = Environment(
                threaded_executor=False,
                threads_per_worker=1,
                worker_context=worker_context
            )

            queues.response.put({
                "type": "STARTUP_DONE",
                "worker_id": worker_idx,
            })

        return worker_loop(queues, work_mem, worker_idx, env)
    except Exception as e:
        queues.response.put({
            "type": "ERROR",
            "worker_id": worker_idx,
            "error": e,
            "exception": e,
        })
        # drain, close and join queues:
        queues.request.close()
        queues.response.close(drain=False)
        queues.message.close(drain=False)
        sys.exit(1)


class PipelinedWorkerContext(WorkerContext):
    """
    A context object that is made available to the Partition
    for custom communication to its matching DataSet class
    (currently uni-directional)
    """
    def __init__(self, queue: "WorkerQueue", msg_queue: "WorkerQueue"):
        self._queue = queue
        self._msg_queue = msg_queue

    def get_worker_queue(self) -> WorkerQueue:
        return self._queue

    def signal(self, ident: str, topic: str, msg_dict: dict[str, Any]):
        msg_dict.update({'ident': ident})
        self._msg_queue.put((topic, msg_dict))


ResultT = Generator[tuple[Any, TaskProtocol], None, None]
ResultWithID = Generator[tuple[Any, TaskProtocol, int], None, None]


def _order_results(results_in: ResultWithID) -> ResultT:
    """
    Order the `results_in` generator by the result id, yielding ordered results.
    Requires indexes to be without gaps.
    """
    last_sent_id = -1
    # for results that are received out-of-order, keep sorted:
    result_stack: list[tuple[Any, TaskProtocol, int]] = []
    span = trace.get_current_span()
    for result, task, task_id in results_in:
        if task_id == last_sent_id + 1:
            span.add_event("_order_results.yield", {
                'task_id': task_id,
            })
            yield (result, task)
            last_sent_id = task_id
        else:
            span.add_event("_order_results.postpone", {
                "expect": last_sent_id + 1,
                "is": task_id,
            })
            result_stack = sorted(
                result_stack + [(result, task, task_id)],
                key=lambda x: x[-1]
            )
        # top of the stack looks good, yield as long as it matches:
        while len(result_stack) > 0 and result_stack[0][2] == last_sent_id + 1:
            res = result_stack.pop(0)
            span.add_event("_order_results.yield", {
                'task_id': res[2],
            })
            yield res[0], res[1]
            last_sent_id = res[2]

    for result, task, task_id in result_stack:
        if task_id != last_sent_id + 1:
            raise RuntimeError(
                f"missing tasks? end of result but next id on result_stack is "
                f"{task_id}, was expecting {last_sent_id + 1}"
            )
        span.add_event("_order_results.yield")
        yield (result, task)
        last_sent_id = task_id


def _make_spec(
    cpus: Union[int, Iterable[int]],
    cudas: Union[int, Iterable[int]],
    has_cupy: bool = False,
) -> list[WorkerSpec]:
    """
    Takes the output of :func:`libertem.utils.devices.detect`
    and makes a plan for starting workers on them.

    Parameters
    ----------

    cpus: int | Iterable
        Iterable of integer CPU identifiers or an integer number of workers to create.
        If pinning is enabled, each worker processe is pinned to one of these identifiers,
        as accepted by :func:`python:os.sched_setaffinity`. Pinning is currently only
        supported on platforms that implement :func:`python:os.sched_setaffinity`.

    cudas: int | Iterable
        Interable of CUDA device identifiers for which workers should be started or
        an integer number of GPU workers to create across the available devices.
        Identifiers can be repeated to start multiple workers per GPU, which can
        result in better device utilization.

    has_cupy
        Is cupy available?
    """
    spec = []
    worker_idx = 0

    if isinstance(cpus, int):
        cpus = tuple(range(cpus))

    cudas = assign_cudas(cudas)

    grouped_cudas = itertools.groupby(cudas, lambda x: x)
    for device_id, group in grouped_cudas:
        for i in range(len(list(group))):
            spec.append(WorkerSpec(
                name=f"cuda-{device_id}-{i}",
                device_id=device_id,
                device_kind="CUDA",
                worker_idx=worker_idx,
                has_cupy=has_cupy,
            ))
            worker_idx += 1

    for device_id in cpus:
        spec.append(WorkerSpec(
            name=f"cpu-{device_id}",
            device_id=device_id,
            device_kind="CPU",
            worker_idx=worker_idx,
            has_cupy=False,
        ))
        worker_idx += 1

    return spec


def _raise_from_msg(msg: dict, err_prefix: str):
    if "exception" in msg:
        raise msg["exception"]
    else:
        raise RuntimeError(f"{err_prefix}: {msg['error']}")


def _inspect_startup(msg, span):
    if msg["type"] == "ERROR":
        _raise_from_msg(msg, "error on startup")
    if msg["type"] != "STARTUP_DONE":
        raise RuntimeError(
            f"unknown message type {msg['type']}, expected STARTUP_DONE"
        )
    span.add_event("worker startup done", {"worker_id": msg["worker_id"]})


class PipelinedExecutor(BaseJobExecutor):
    """
    Multi-process pipelined executor. Useful for live processing using
    `LiberTEM-live <https://libertem.github.io/LiberTEM-live/>`_
    if your processing function is not able to keep up with the incoming data
    stream in a single process, but also works for offline processing.

    Parameters
    ----------
    spec
        Specification for the worker processes - can be generated
        by :meth:`make_spec`.

    pin_workers
        Pin each CPU worker to a specific CPU, as defined by :func:`python:os.sched_setaffinity`.
        Only works on OSes that implement :func:`python:os.sched_setaffinity`.

    startup_timeout
        Startup of the executor is cancelled if it doesn't finish within
        this limit (in detail: each worker's startup time is limited by this timeout).
        In seconds.

    cleanup_timeout
        When cleaning up using :meth:`close`, give up after
        this limit. In seconds.

    early_setup
        Callable that will be run as early as possible on each worker process.
        Useful for custom warmup code or testing.

    Note
    ----
    This executor is not thread-safe - concurrent calls into :meth:`run_tasks` or
    :meth:`run_function` are not supported.
    """
    def __init__(
        self,
        spec: Optional[list[WorkerSpec]] = None,
        pin_workers: bool = True,
        startup_timeout: float = 30.0,
        cleanup_timeout: float = 10.0,
        early_setup: Optional[Callable] = None,
    ) -> None:
        self._pin_workers = pin_workers
        if spec is None:
            spec = self._default_spec()
        self._spec = spec
        self._closed = True
        self._early_setup = early_setup

        # for testing via monkeypatching, for example:
        self._worker_selector: Optional[WorkerSelector] = None

        # timeout for cleanup, either from exception or when joining processes
        self._cleanup_timeout = cleanup_timeout

        # timeout for starting a single worker
        self._startup_timeout = startup_timeout

        # keep this at the bottom:
        self._pool = self._start_pool()

    def _start_pool(self) -> WorkerPool:
        with tracer.start_as_current_span("PipelinedExecutor.start_pool") as span:
            pool = WorkerPool(
                worker_fn=functools.partial(
                    pipelined_worker,
                    pin=self._pin_workers,
                    early_setup=self._early_setup,
                ),
                spec=self._spec,
            )
            # if any processes are already dead here, raise an exception so we
            # don't have to run into a timeout below:
            pool.assert_all_alive()

            warn_time = 5.0
            check_interval = 0.1
            warn_deadline = time.monotonic() + warn_time
            startup_deadline = time.monotonic() + self._startup_timeout
            warned = False

            num_started = 0

            while num_started < pool.size:
                pool.assert_all_alive()
                try:
                    with pool.response_queue.get(timeout=check_interval) as (msg, _):
                        _inspect_startup(msg, span)
                        num_started += 1
                        continue
                except WorkerQueueEmpty:
                    if time.monotonic() > startup_deadline:
                        pool.assert_all_alive()
                        pool.kill()
                        # break possibly confusing exception chain using "from None":
                        raise RuntimeError(
                            f"Timeout while starting workers, might need to increase "
                            f"`startup_timeout` (is {self._startup_timeout}s)"
                        ) from None
                    if time.monotonic() > warn_deadline and not warned:
                        warnings.warn('Slow worker startup, please be patient...', RuntimeWarning)
                        warned = True
            for qp in pool.workers:
                qp.queues.request.put({
                    "type": "WARMUP",
                    "span_context": span.get_span_context(),
                })
            # set here, so we don't try to close the pool if it doesn't exist
            self._closed = False
            return pool

    def _restart_pool(self):
        self._pool.kill()
        self._pool = self._start_pool()

    @classmethod
    def _default_spec(cls):
        from libertem.utils.devices import detect
        detected = detect()
        return _make_spec(**detected)

    @classmethod
    def make_local(cls, **kwargs):
        """
        Create a :code:`PipelinedExecutor` with the default spec.
        """
        spec = cls._default_spec()
        return cls(spec=spec, **kwargs)

    @classmethod
    def make_spec(
        cls,
        cpus: Union[int, Iterable[int]],
        cudas: Union[int, Iterable[int]],
        has_cupy: bool = False,
    ):
        return _make_spec(cpus=cpus, cudas=cudas, has_cupy=has_cupy)
    make_spec.__doc__ = _make_spec.__doc__

    def _validate_worker_state(self):
        if not self._pool.all_alive():
            raise RuntimeError("some workers are stopped, cannot continue")

    def _run_tasks_inner(
        self,
        tasks: Iterable[TaskProtocol],
        params_handle: Any,
        cancel_id: Any,
        task_comm_handler: "TaskCommHandler",
    ) -> ResultWithID:
        # In theory, `in_flight` could be calculated from `id_to_task`, but in case
        # of exceptions, it becomes a bit harder to keep attribution of messages to
        # tasks, which is why we have a separate counter for now.
        in_flight = 0
        id_to_task: dict[int, TaskProtocol] = {}
        tasks_uuid = str(uuid.uuid4())

        all_workers = self.get_available_workers()
        scheduler = Scheduler(all_workers=all_workers)
        pool_info_for_worker = dict(zip(all_workers, self._pool.workers))
        selector: WorkerSelector

        # feature detect: can we access the queue size? support for this can be
        # missing depending on platform
        if self._worker_selector is None:
            try:
                self._pool.workers[0].queues.request.size()
                selector = _select_by_queue_size
            except NotImplementedError:
                # fall back to simple round robin worker selection
                selector = _select_by_round_robin
        else:
            selector = self._worker_selector

        try:
            self._validate_worker_state()
            task_comm_handler.start()
            span = trace.get_current_span()
            span_context = span.get_span_context()

            def yield_result_if_found(block, timeout):
                nonlocal in_flight
                try:
                    with self._pool.response_queue.get(block=block, timeout=timeout) as (result, _):
                        if result.get('uuid') != tasks_uuid:
                            # mismatch, log and ignore:
                            logger.warning(
                                "mismatched result, ignoring: %s != %s",
                                result.get('uuid'), tasks_uuid,
                            )
                            return
                        in_flight -= 1
                        if result["type"] == "ERROR":
                            _raise_from_msg(result, "failed to run tasks")
                        result_task_id = result["task_id"]
                        yield (result["result"], id_to_task[result_task_id], result_task_id)
                        del id_to_task[result_task_id]
                        if len(id_to_task) != in_flight:
                            raise RuntimeError(
                                "state mismatch; `id_to_task` mapping should match `in_flight`"
                            )
                except WorkerQueueEmpty:
                    self._validate_worker_state()

            with tracer.start_as_current_span('schedule_tasks') as span:
                for task_idx, task in enumerate(tasks):
                    # important: call `schedule_task` before incrementing number of
                    # in-flight tasks, otherwise, if we run into an exception, we
                    # will wait for a response that never comes.
                    # select a worker for `task` according to `selector`:
                    _worker_idx, worker_queues = schedule_task(
                        task_idx,
                        task,
                        self._pool,
                        scheduler,
                        pool_info_for_worker,
                        selector=selector,
                    )
                    worker_queues.request.put({
                        "type": "RUN_TASK",
                        "uuid": tasks_uuid,
                        "task": cloudpickle.dumps(task),
                        "task_id": task_idx,
                        "params_handle": params_handle,
                        "span_context": span_context,
                    })
                    span.add_event('RUN_TASK', {
                        'task_id': task_idx,
                    })
                    in_flight += 1
                    id_to_task[task_idx] = task
                    if len(id_to_task) != in_flight:
                        raise RuntimeError(
                            "state mismatch; `id_to_task` mapping should match `in_flight`"
                        )

                    # FIXME: semantics of this - is this enough?
                    # does it matter if this is enough? we can change it in the future if not
                    # could be: the function returns once it has forwarded
                    # all the data necessary for the given task,
                    # (or, in the offline case, immediately)
                    try:
                        task_comm_handler.handle_task(task, worker_queues.request)
                    except JobCancelledError:
                        worker_queues.request.put({
                            "type": "END_TASK",
                            "task_id": task_idx,
                            "params_handle": params_handle,
                            "span_context": span_context,
                        })
                        raise

                    # NOTE: sentinel message; in case of errors, the worker
                    # needs to discard the data from the queue until it receives
                    # this message:
                    worker_queues.request.put({
                        "type": "END_TASK",
                        "task_id": task_idx,
                        "params_handle": params_handle,
                        "span_context": span_context,
                    })

                    yield from yield_result_if_found(block=False, timeout=None)

            # at the end, block to get the remaining results:
            while in_flight > 0:
                yield from yield_result_if_found(block=True, timeout=0.1)
        except JobCancelledError:
            # don't drain here, as the next acquisition could start
            # very soon; instead, we just ignore the mismatched
            # responses in the next run
            raise
        except Exception as e:
            # In case of an exception, we need to drain the response queue,
            # so the next `run_tasks` call isn't polluted by old responses.
            # -> just like in the happy case, we need to wait for responses
            # for all in-flight tasks. In case the error happened between incrementing
            # `in_flight` and actually sending the task to the queue, we should
            # have a timeout here to not wait infinitely long.
            try:
                self._drain_response_queue(in_flight=in_flight)
            except RuntimeError as e2:
                raise e2 from e
            # if from a worker, this is the first exception that got put into the queue
            raise
        finally:
            task_comm_handler.done()

    def _drain_response_queue(self, in_flight: int) -> None:
        """
        Drain response queue and log any errors returned from workers

        Parameters
        ----------
        in_flight : int
            The number of requests that are still in flight
        """
        t0 = time.time()
        while in_flight > 0:
            try:
                timeout = max(0.010, self._cleanup_timeout - (time.time() - t0))
                with self._pool.response_queue.get(
                    timeout=timeout,
                ) as (result, _):
                    t0 = time.time()
                    in_flight -= 1
                    # we only raise the first exception; log the others here:
                    if result["type"] == "ERROR":
                        logger.error(f"Error response from worker: {result['error']}")
            except WorkerQueueEmpty:
                # kill and restart workers:
                self._restart_pool()
                raise RuntimeError(
                    f'Worker or queue presumably in a bad state, lost {in_flight} in-flight tasks.'
                )

    def run_tasks(
        self,
        tasks: Iterable[TaskProtocol],
        params_handle: Any,
        cancel_id: Any,
        task_comm_handler: "TaskCommHandler",
    ) -> ResultT:
        with tracer.start_as_current_span("PipelinedExecutor.run_tasks"):
            with task_comm_handler.monitor(self._pool.message_queue):
                yield from _order_results(self._run_tasks_inner(
                    tasks, params_handle, cancel_id, task_comm_handler,
                ))

    def get_available_workers(self) -> WorkerSet:
        return WorkerSet([
            Worker(
                name=worker_info.spec["name"],
                host="localhost",
                resources=_resources_for_spec(worker_info.spec),
                nthreads=1,
            )
            for worker_info in self._pool.workers
        ])

    def close(self):
        with tracer.start_as_current_span("PipelinedExecutor.close") as span:
            if self._closed:
                return
            for idx, worker_info in enumerate(self._pool.workers):
                span.add_event("sending SHUTDOWN", {"idx": idx})
                worker_info.queues.request.put({
                    "type": "SHUTDOWN",
                    "span_context": span.get_span_context(),
                })
                span.add_event("SHUTDOWN sent", {"idx": idx})
                while True:
                    try:
                        with worker_info.queues.response.get(block=False) as msg:
                            logger.warning(f"got message on close: {msg[0]}")
                    except WorkerQueueEmpty:
                        break
                worker_info.process.join(timeout=self._cleanup_timeout)
                if worker_info.process.exitcode is None:
                    self._pool.kill_worker(worker_info)
                worker_info.queues.request.close(force=True)
            self._pool.close_resp_queue()
            self._pool.close_mesg_queue()
            self._closed = True

    def __del__(self):
        # `self` may be already partially garbage-collected; only close
        # if "enough" of `self` still exists:
        if hasattr(self, '_closed') and not self._closed:
            self.close()

    def _run_function(self, fn: Callable[..., T], worker_idx, *args, **kwargs) -> T:
        self._validate_worker_state()
        qs = self._pool.get_worker_queues(worker_idx)
        f = functools.partial(fn, *args, **kwargs)
        pickled = cloudpickle.dumps(f)
        qs.request.put({
            "type": "RUN_FUNCTION",
            "fn": pickled,
            "span_context": trace.get_current_span().get_span_context(),
        })
        # FIXME: timeout?
        with qs.response.get() as (response, _):
            if response["type"] == "ERROR":
                if "exception" in response:
                    raise response["exception"]
                else:
                    raise RuntimeError(f"failed to run function: {response['error']}")
            if not response["type"] == "RUN_FUNCTION_RESULT":
                raise RuntimeError(f"invalid response type: {response['type']}")
            result: T = response["result"]
            return result

    def run_function(self, fn: Callable[..., T], *args, **kwargs) -> T:
        # FIXME: this is not concurrency-safe currently! beware!
        with tracer.start_as_current_span("PipelinedExecutor.run_function"):
            return self._run_function(fn, 0, *args, **kwargs)

    def run_each_worker(self, fn, *args, **kwargs):
        # FIXME: not as fast as it could be, but also not really perf-sensitive?
        result = {}
        for idx, worker_info in enumerate(self._pool.workers):
            result[worker_info.spec["name"]] = self._run_function(
                fn, worker_idx=idx, *args, **kwargs
            )
        return result

    def run_each_host(self, fn, *args, **kwargs):
        return {"localhost": self.run_function(fn, *args, **kwargs)}

    @contextlib.contextmanager
    def scatter(self, obj):
        self._validate_worker_state()
        key = str(uuid.uuid4())
        with tracer.start_as_current_span("scatter_put_to_workers"):
            for worker_info in self._pool.workers:
                worker_info.queues.request.put({
                    "type": "SCATTER",
                    "key": key,
                    "value": obj,
                })
        try:
            yield key
        finally:
            if not self._closed:
                for worker_info in self._pool.workers:
                    worker_info.queues.request.put({
                        "type": "DELETE",
                        "key": key,
                    })

    def scatter_update(self, handle: str, obj):
        self._validate_worker_state()
        for worker_info in self._pool.workers:
            worker_info.queues.request.put({
                "type": "SCATTER_UPDATE",
                "key": handle,
                "value": obj,
            })

    def scatter_update_patch(self, handle: str, patch):
        self._validate_worker_state()
        for worker_info in self._pool.workers:
            worker_info.queues.request.put({
                "type": "SCATTER_UPDATE_PATCH",
                "key": handle,
                "patch": patch,
            })

    def map(self, fn, iterable):
        # FIXME: replace with efficient impl if needed
        with tracer.start_as_current_span("PipelinedExecutor.map"):
            return [
                self._run_function(fn=lambda: fn(item), worker_idx=0)
                for item in iterable
            ]



from libertem.common import Shape


def test_shape_get_nav():
    s = Shape((16, 16, 128, 128), sig_dims=2)
    assert tuple(s.nav) == (16, 16)


def test_shape_get_sig():
    s = Shape((16, 16, 128, 128), sig_dims=2)
    assert tuple(s.sig) == (128, 128)


def test_shape_get_size():
    s = Shape((16, 16, 128, 128), sig_dims=2)
    assert s.size == 16 * 16 * 128 * 128


def test_size_zero():
    s = Shape((), sig_dims=0)
    assert s.size == 0


def test_size_nav_zero():
    s = Shape((128, 128), sig_dims=2)
    assert s.nav.size == 0


def test_shape_flatten_nav():
    s = Shape((16, 16, 128, 128), sig_dims=2)
    assert tuple(s.flatten_nav()) == (16 * 16, 128, 128)


def test_shape_flatten_sig():
    s = Shape((16, 16, 128, 128), sig_dims=2)
    assert tuple(s.flatten_sig()) == (16, 16, 128 * 128)


def test_shape_getitem():
    s = Shape((16, 16, 128, 128), sig_dims=2)
    assert s[:2] == (16, 16)
    assert s[0] == 16


def test_shape_len():
    s = Shape((16, 16, 128, 128), sig_dims=2)
    assert len(s) == 4


def test_shape_to_tuple():
    s = Shape((16, 16, 128, 128), sig_dims=2)
    assert s.to_tuple() == (16, 16, 128, 128)


def test_shape_repr():
    s = Shape((16, 16, 128, 128), sig_dims=2)
    assert repr(s) == "(16, 16, 128, 128)"


def test_shape_eq_1():
    s1 = Shape((16, 16, 128, 128), sig_dims=2)
    s2 = Shape((16, 16, 128, 128), sig_dims=2)
    assert s1 == s2


def test_shape_eq_2():
    s1 = Shape((16, 16, 128, 128), sig_dims=2)
    s2 = Shape((16, 16, 128, 128), sig_dims=3)
    assert s1 != s2


def test_shape_eq_3():
    s1 = Shape((16, 16, 128, 128), sig_dims=2)
    s2 = Shape((16 * 16, 128, 128), sig_dims=2)
    assert s1 != s2 and s1.sig.dims == s2.sig.dims


def test_shape_eq_4():
    s1 = Shape((17, 16, 128, 128), sig_dims=2)
    s2 = Shape((16, 16, 128, 128), sig_dims=2)
    assert s1 != s2 and s1.sig.dims == s2.sig.dims


def test_shape_add_1():
    s = Shape((12, 13, 14, 15), sig_dims=2)
    s_add = (1, 2) + s
    assert tuple(s_add) == (12, 13, 1, 2, 14, 15) and s_add.sig.dims == 2


def test_shape_add_2():
    s = Shape((12, 13, 14, 15), sig_dims=2)
    s_add = s + (1, 2)
    assert tuple(s_add) == (12, 13, 14, 15, 1, 2) and s_add.sig.dims == 4


def test_can_hash():
    shape = (12, 13, 14, 15)
    sd1 = Shape(shape, sig_dims=1)
    mapping = {sd1: 55}
    sd2 = Shape(shape, sig_dims=2)
    mapping[sd2] = 66
    assert len(mapping) == 2
    mapping.pop(sd1)
    assert sd1 not in mapping.keys()



from enum import Enum
from typing_extensions import Protocol, TypedDict
from typing import Union
from sparseconverter import (
    CUDA, CUPY, CUPY_SCIPY_COO, CUPY_SCIPY_CSC, CUPY_SCIPY_CSR, NUMPY,
    SCIPY_COO, SCIPY_CSC, SCIPY_CSR, SPARSE_COO, SPARSE_DOK, SPARSE_GCXS,
    CPU_BACKENDS, CUDA_BACKENDS, CUPY_BACKENDS, SPARSE_BACKENDS, DENSE_BACKENDS,
    ND_BACKENDS, D2_BACKENDS,
)

import numpy as np


# markers for special values:
class TileDepthEnum(Enum):
    TILE_DEPTH_DEFAULT = object()


class TileSizeEnum(Enum):
    TILE_SIZE_BEST_FIT = object()


class TilingPreferences(TypedDict):
    depth: Union[int, TileDepthEnum]
    total_size: Union[float, int]


class UDFMethod(Enum):
    TILE = 'tile'
    FRAME = 'frame'
    PARTITION = 'partition'


class UDFProtocol(Protocol):
    '''
    Parts of the UDF interface required for MIT code in LiberTEM
    '''
    USE_NATIVE_DTYPE = bool  #: Neutral element for type conversion
    TILE_SIZE_BEST_FIT = TileSizeEnum.TILE_SIZE_BEST_FIT  #: Suggest using recommended tile size
    TILE_SIZE_MAX = np.inf  #: Suggest using maximum tile size
    TILE_DEPTH_DEFAULT = TileDepthEnum.TILE_DEPTH_DEFAULT  #: Suggest using recommended tile depth
    TILE_DEPTH_MAX = np.inf  #: Suggest using maximum tile depth
    BACKEND_NUMPY = NUMPY  #: NumPy array
    BACKEND_CUPY = CUPY  #: CuPy array
    BACKEND_CUDA = CUDA  #: NumPy array, but run on CUDA device class
    BACKEND_SPARSE_COO = SPARSE_COO  #: sparse.COO array
    BACKEND_SPARSE_GCXS = SPARSE_GCXS  #: sparse.GCXS array
    BACKEND_SPARSE_DOK = SPARSE_DOK  #: sparse.DOK array -- not recommended since slow!
    BACKEND_SCIPY_COO = SCIPY_COO  #: scipy.sparse.coo_matrix
    BACKEND_SCIPY_CSR = SCIPY_CSR  #: scipy.sparse.csr_matrix
    BACKEND_SCIPY_CSC = SCIPY_CSC  #: scipy.sparse.csc_matrix
    BACKEND_CUPY_SCIPY_COO = CUPY_SCIPY_COO  #: cupyx.scipy.sparse.coo_matrix
    BACKEND_CUPY_SCIPY_CSR = CUPY_SCIPY_CSR  #: cupyx.scipy.sparse.csr_matrix
    BACKEND_CUPY_SCIPY_CSC = CUPY_SCIPY_CSC  #: cupyx.scipy.sparse.csc_matrix
    # Excludes sparse.DOK, numpy.matrix and CUDA, prefers scipy.sparse and GPU
    # Deprioritizes sparse.pydata.org due to their high call overhead
    #: Tuple with all backends in suggested priority
    BACKEND_ALL = (
        CUPY_SCIPY_CSR, CUPY_SCIPY_CSC, CUPY_SCIPY_COO,
        SCIPY_CSR, SCIPY_CSC, SCIPY_COO,
        CUPY, NUMPY,
        SPARSE_COO, SPARSE_GCXS,
    )

    CPU_BACKENDS = CPU_BACKENDS  #: Set of backends that run on device class CPU
    CUDA_BACKENDS = CUDA_BACKENDS  #: Set of backends that run on device class CUDA
    CUPY_BACKENDS = CUPY_BACKENDS  #: Set of backends that use CuPy, subset of class CUDA
    SPARSE_BACKENDS = SPARSE_BACKENDS  #: Set of backends that are sparse arrays
    DENSE_BACKENDS = DENSE_BACKENDS  #: Set of backends that are dense arrays
    ND_BACKENDS = ND_BACKENDS  #: Set of backends that support n-dimensional arrays
    D2_BACKENDS = D2_BACKENDS  #: Set of backends that only support two-dimensional arrays

    UDF_METHOD = UDFMethod  #: Enum of process_ methods accepted by the UDF interface

    def get_method() -> UDFMethod:
        raise NotImplementedError()

    def get_tiling_preferences() -> TilingPreferences:
        raise NotImplementedError()



from typing import Any, Optional, Union, TYPE_CHECKING, Generic, TypeVar
from collections.abc import Iterable
from typing_extensions import Literal
import mmap
import math
import functools
from contextlib import contextmanager
import collections
import itertools

try:
    from itertools import batched
except ImportError:
    def batched(iterable, n):
        # batched('ABCDEFG', 3) --> ABC DEF G
        if n < 1:
            raise ValueError('n must be at least one')
        it = iter(iterable)
        while batch := tuple(itertools.islice(it, n)):
            yield batch

import numba
import numpy as np

from libertem.common.math import prod, count_nonzero, flat_nonzero
from libertem.common.slice import Slice, Shape
from .backend import get_use_cuda

if TYPE_CHECKING:
    from numpy import typing as nt

BufferKind = Literal['nav', 'sig', 'single']
BufferLocation = Optional[Literal['device']]
BufferUse = Literal['private', 'result_only']
BufferSize = Union[int, tuple[int, ...]]


def _alloc_aligned(size: int, blocksize: int = 4096) -> mmap.mmap:
    # round up to (default 4k) blocks:
    blocks = math.ceil(size / blocksize)

    # flags are by default MAP_SHARED, which has to be set
    # to prevent possible corruption (see open(2)). If you
    # are adding flags here, make sure to include MAP_SHARED
    # (and check for windows compat)
    buf = mmap.mmap(-1, blocksize * blocks)

    # FIXME: if `blocksize` > PAGE_SIZE, we may have to align the start
    # address here, too.

    return buf


def bytes_aligned(size: int) -> memoryview:
    buf = _alloc_aligned(size)
    # _alloc_aligned may give us more memory (for alignment reasons), so crop it off the end:
    return memoryview(buf)[:size]


def empty_aligned(size: BufferSize, dtype: "nt.DTypeLike") -> np.ndarray:
    size_flat = prod(size)
    dtype = np.dtype(dtype)
    buf = _alloc_aligned(dtype.itemsize * size_flat)
    # _alloc_aligned may give us more memory (for alignment reasons), so crop it off the end:
    npbuf: np.ndarray = np.frombuffer(buf, dtype=dtype)[:size_flat]
    return npbuf.reshape(size)


def zeros_aligned(size: BufferSize, dtype: "nt.DTypeLike") -> np.ndarray:
    if dtype == object or prod(size) == 0:
        res = np.zeros(size, dtype=dtype)
    else:
        res = empty_aligned(size, dtype)
        res[:] = 0
    return res


# FIXME: type annotation for cupy.ndarray without importing?
def to_numpy(a: Union[np.ndarray, Any]) -> np.ndarray:
    # .. versionadded:: 0.6.0
    cuda_device = get_use_cuda()
    if isinstance(a, np.ndarray):
        return a
    elif cuda_device is not None:
        # Try to avoid importing unless necessary
        import cupy
        if isinstance(a, cupy.ndarray):
            return cupy.asnumpy(a)
    # Falling through
    raise TypeError(f"I don't know how to convert {type(a)} here.")


def reshaped_view(a: np.ndarray, shape):
    '''
    Like :meth:`numpy.ndarray.reshape`, just guaranteed to
    return a view or throw an :class:`AttributeError` if
    no view can be created.

    .. versionadded:: 0.5.0

    Parameters
    ----------

    a : numpy.ndarray
        Array to create a view of
    shape : tuple
        Shape of the view to create

    Returns
    -------

    view : numpy.ndarray
        View into :code:`a` with shape :code:`shape`

    '''
    res = a.view()
    res.shape = shape
    return res


def disjoint(sl: Slice, slices: Iterable[Slice]):
    return all(sl.intersection_with(s2).is_null() for s2 in slices)


class BufferPool:
    """
    allocation pool for explicitly re-using (aligned) allocations
    """
    def __init__(self):
        self._buffers = collections.defaultdict(list)

    @contextmanager
    def zeros(self, size, dtype, alignment=4096):
        if dtype == object or prod(size) == 0:
            yield np.zeros(size, dtype=dtype)
        else:
            with self.empty(size, dtype, alignment) as res:
                res[:] = 0
                yield res

    @contextmanager
    def empty(self, size, dtype, alignment=4096):
        size_flat = prod(size)
        dtype = np.dtype(dtype)
        with self.bytes(dtype.itemsize * size_flat, alignment) as buf:
            # self.bytes may give us more memory (for alignment reasons), so
            # crop it off the end:
            npbuf = np.frombuffer(buf, dtype=dtype)[:size_flat]
            yield npbuf.reshape(size)

    @contextmanager
    def bytes(self, size, alignment=4096):
        buf = self.checkout_bytes(size, alignment)
        yield buf
        self.checkin_bytes(size, alignment, buf)

    def checkout_bytes(self, size, alignment):
        buffers = self._buffers[(size, alignment)]
        try:
            buf = buffers.pop()
        except IndexError:
            buf = _alloc_aligned(size, blocksize=alignment)
        return buf

    def checkin_bytes(self, size, alignment, buf):
        self._buffers[(size, alignment)].insert(0, buf)


class ManagedBuffer:
    """
    Allocate `size` bytes from `pool`, and return them to the pool once we are GC'd
    """
    def __init__(self, pool, size, alignment):
        self.pool = pool
        self.buf = pool.checkout_bytes(size, alignment)
        self.size = size
        self.alignment = alignment

    def __del__(self):
        self.pool.checkin_bytes(self.size, self.alignment, self.buf)


T = TypeVar('T', bound=np.generic)


class InvalidMaskError(Exception):
    """
    The mask is not compatible, raised for example when the shape of
    the mask doesn't match the data.
    """
    pass


class ArrayWithMask(Generic[T]):
    """
    An opaque type representing an array together with a
    mask (for example, defining a region in the array
    where there are valid entries)

    This is meant for usage in :meth:`UDF.get_results`,
    and you can use the convenience method :meth:`UDF.with_mask`
    to instantiate it.
    """

    def __init__(self, arr: T, mask: Union[np.ndarray, bool]):
        """
        :meta private:
        """
        if isinstance(mask, bool):
            mask = np.array([mask])
        try:
            mask = np.broadcast_to(mask, arr.shape)
        except ValueError:
            raise InvalidMaskError(
                f"`arr` and `mask` must have compatible shapes "
                f"(arr.shape={arr.shape} vs mask.shape={mask.shape})"
            )
        if mask.dtype != np.dtype('bool'):
            raise InvalidMaskError(
                f"`mask` should have `dtype=bool` (have {mask.dtype})"
            )
        self._arr = arr
        self._mask = mask

    @property
    def mask(self) -> np.ndarray:
        return np.broadcast_to(self._mask, self._arr.shape)

    @property
    def arr(self) -> T:
        return self._arr


def get_inner_slice(arr: np.ndarray, axis: int = 0) -> tuple[slice, ...]:
    """
    Get a slice into `arr` across `axis` where the values on all other axes are non-zero.
    This will be the first contiguous slice, not necessarily the largest.

    :meta private:
    """
    ax_min = arr.shape[axis]
    ax_max = 0
    state = 0

    other_axes = tuple([
        i for i in range(arr.ndim)
        if i != axis
    ])
    non_zero = np.all(arr != 0, axis=other_axes)

    # find the first contiguous slice:
    for i in range(non_zero.shape[0]):
        if non_zero[i]:
            if state == 0:
                state = 1
                ax_min = i
                ax_max = i
            elif state == 1:
                ax_max = i
        else:
            if state == 1:
                break
    slice_ = tuple([
        slice(ax_min, ax_max + 1) if dim == axis else slice(None, None, None)
        for dim in range(arr.ndim)
    ])
    return slice_


@numba.njit(cache=True)
def get_bbox_2d(arr, eps=1e-8) -> tuple[int, ...]:
    """
    :meta private:
    """
    xmin = arr.shape[1]
    ymin = arr.shape[0]
    xmax = 0
    ymax = 0

    for y in range(arr.shape[0]):
        for x in range(arr.shape[1]):
            value = arr[y, x]
            if abs(value) < eps:
                continue
            # got a non-zero value, update indices
            if x < xmin:
                xmin = x
            if x > xmax:
                xmax = x
            if y < ymin:
                ymin = y
            if y > ymax:
                ymax = y
    return int(ymin), int(ymax), int(xmin), int(xmax)


def get_bbox(arr: np.ndarray) -> tuple[int, ...]:
    """
    :meta private:
    """
    if arr.ndim == 2:
        # 4x faster specialization in numba for 2D:
        return get_bbox_2d(arr)
    else:
        # from https://stackoverflow.com/a/31402351/540644
        N = arr.ndim
        out: list[int] = []
        for ax in itertools.combinations(reversed(range(N)), N - 1):
            nonzero = np.any(arr, axis=ax)
            out.extend(np.where(nonzero)[0][[0, -1]])
        return tuple(out)


def get_bbox_slice(arr: np.ndarray) -> tuple[slice, ...]:
    """
    :meta private:
    """
    bbox = get_bbox(arr)
    res = []
    for pair in batched(bbox, 2):
        res.append(slice(pair[0], pair[1] + 1, None))
    return tuple(res)


class BufferWrapper:
    """
    Helper class to automatically allocate buffers, either for partitions or
    the whole dataset, and create views for partitions or single frames.

    This is used as a helper to allow easy merging of results without needing
    to manually handle indexing.

    Usually, as a user, you only need to instantiate this class, specifying `kind`,
    `dtype` and sometimes `extra_shape` parameters. Most methods are meant to be called
    from LiberTEM-internal code, for example the UDF functionality.

    This class is array_like, so you can directly use it, for example, as argument
    for numpy functions.

    .. versionchanged:: 0.6.0
        Add option to specify backend, for example CuPy

    Parameters
    ----------
    kind : "nav", "sig" or "single"
        The abstract shape of the buffer, corresponding either to the navigation
        or the signal dimensions of the dataset, or a single value.

    extra_shape : optional, tuple of int or a Shape object
        You can specify additional dimensions for your data. For example, if
        you want to store 2D coords, you would specify (2,) here.
        For a Shape object, sig_dims is discarded and the entire shape is used.

    dtype : string or numpy dtype
        The dtype of this buffer

    where : string or None
        :code:`None` means NumPy array, :code:`device` to use a back-end specified
        in :meth:`allocate`.

        .. versionadded:: 0.6.0

    use : "private", "result_only" or None
        If you specify :code:`"private"` here, the result will only be made available
        to internal functions, like :meth:`process_frame`, :meth:`merge` or
        :meth:`get_results`. It will not be available to the user of the UDF, which means
        you can use this to hide implementation details that are likely to change later.

        Specify :code:`"result_only"` here if the buffer is only used in :meth:`get_results`,
        this means we don't have to allocate and return it on the workers without actually
        needing it.

        :code:`None` means the buffer is used both as a final and intermediate result.

        .. versionadded:: 0.7.0

    """
    def __init__(
        self,
        kind: Literal['nav', 'sig', 'single'],
        extra_shape: tuple[int, ...] = (),
        dtype="float32",
        where: Union[None, Literal['device']] = None,
        use: Optional[Literal['private', 'result_only']] = None,
    ) -> None:
        self._extra_shape = tuple(extra_shape)
        self._kind = kind
        self._dtype = np.dtype(dtype)
        self._where = where
        self._data: Optional[np.ndarray] = None
        # set to True if the data coords are global ds coords
        self._data_coords_global = False
        self._shape = None
        self._ds_shape: Optional[Shape] = None
        self._ds_partitions = None
        self._roi = None
        self._roi_is_zero = None
        self._contiguous_cache = dict()
        self.use = use

    def set_roi(self, roi):
        """
        :meta private:
        """
        if roi is not None:
            roi = roi.reshape((-1,))
        self._roi = roi

    def add_partitions(self, partitions):
        """
        Add a list of dataset partitions to the buffer such that
        self.allocate() can make use of the structure

        :meta private:
        """
        self._ds_partitions = partitions

    def set_shape_partition(self, partition, roi=None):
        """
        :meta private:
        """
        self.set_roi(roi)
        roi_count = None
        if roi is not None:
            roi_part = self._roi[partition.slice.get(nav_only=True)]
            roi_count = count_nonzero(roi_part)
            assert roi_count <= partition.shape[0]
            assert roi_part.shape[0] == partition.shape[0]
        self._shape = self._shape_for_kind(self._kind, partition.shape, roi_count)
        self._update_roi_is_zero()

    def set_shape_ds(self, dataset_shape: Shape, roi=None):
        """
        :meta private:
        """
        self.set_roi(roi)
        roi_count = None
        if roi is not None:
            roi_count = count_nonzero(self._roi)
        self._shape = self._shape_for_kind(self._kind, dataset_shape.flatten_nav(), roi_count)
        self._update_roi_is_zero()
        self._ds_shape = dataset_shape

    @property
    def shape(self):
        # precondition: _shape_for_kind has been called
        return self._shape

    def _shape_for_kind(self, kind, orig_shape, roi_count=None):
        """
        :meta private:
        """
        if self._kind == "nav":
            if roi_count is None:
                nav_shape = tuple(orig_shape.nav)
            else:
                nav_shape = (roi_count,)
            return nav_shape + self._extra_shape
        elif self._kind == "sig":
            return tuple(orig_shape.sig) + self._extra_shape
        elif self._kind == "single":
            if len(self._extra_shape) > 0:
                return self._extra_shape
            else:
                return (1, )
        else:
            raise ValueError("unknown kind: %s" % kind)

    @property
    def data(self):
        """
        Get the buffer contents in shape that corresponds to the
        original dataset shape. If a ROI is set, embed the result into a new
        array; unset values have NaN value for floating point types,
        False for boolean, 0 for integer types and structs,
        '' for string types and None for objects.

        .. versionchanged:: 0.7.0
            Better initialization values for dtypes other than floating point.
        """
        if self._contiguous_cache:
            raise RuntimeError("Cache is not empty, has to be flushed")
        if self._roi is None or self._kind != 'nav':
            return self._data.reshape(self._shape_for_kind(self._kind, self._ds_shape))
        shape = self._shape_for_kind(self._kind, self._ds_shape)
        if shape == self._data.shape:
            # preallocated and already wrapped
            return self._data
        # Integer types and "void" (structs and such)
        if self.dtype.kind in ('i', 'u', 'V'):
            fill = 0
        # Bytes and Unicode strings
        elif self.dtype.kind in ('S', 'U'):
            fill = ''
        else:
            # 'b' (boolean): False
            # 'f', 'c': NaN
            # 'm', 'M' (datetime, timedelta): NaT
            # 'O' (object): None
            fill = None
        flat_shape_with_extra = (prod(shape) // prod(self._extra_shape),) + self._extra_shape
        wrapper = np.full(flat_shape_with_extra, fill, dtype=self._dtype)
        wrapper[flat_nonzero(self._roi), ...] = self._data
        return wrapper.reshape(shape)

    @property
    def dtype(self):
        """
        Get the declared dtype of this buffer.

        .. versionadded:: 0.7.0
        """
        return self._dtype

    @property
    def raw_data(self) -> Optional[np.ndarray]:
        """
        Get the raw data underlying this buffer, which is flattened and
        may be even filtered to a ROI
        """
        return self._data

    def make_default_mask(
        self,
        valid_nav_mask: np.ndarray,
        dataset_shape: Shape,
        roi: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """
        Create the default valid mask for this buffer kind, given the "upstream" `valid_nav_mask`.

        :meta private:
        """
        roi_count = None
        if roi is not None:
            roi_count = count_nonzero(roi)
        shape = self._shape_for_kind(self._kind, dataset_shape.flatten_nav(), roi_count)
        if self._kind == 'nav':
            # kind=nav by default uses the `valid_nav_mask`, which we need to
            # broadcast to take care of `extra_shape`:
            mask = np.zeros(shape, dtype=bool)
            extra_compat = tuple([
                1
                for _ in range(len(self._extra_shape))
            ])
            valid_nav_mask_compat = valid_nav_mask.reshape(valid_nav_mask.shape + extra_compat)
            mask[:] = valid_nav_mask_compat
        else:
            mask = np.ones(shape, dtype=bool)
        return mask

    @property
    def valid_mask(self) -> np.ndarray:
        """
        Returns a boolean array with a mask that indicates which parts of the
        buffer have valid contents.
        """
        if self._ds_shape is None:
            raise RuntimeError("`valid_mask` called without setting the dataset shape")

        if self._valid_mask is None:
            raise RuntimeError("`valid_mask` must be set")

        if self._kind == 'nav':
            # for `nav`, we need to do some gymnastics to un-flatten the nav
            # part of the shape, and possibly handle `roi`:
            full_shape = tuple(self._ds_shape.nav) + self._extra_shape
            if self._roi is not None:
                flat_shape = tuple(self._ds_shape.flatten_nav().nav) + self._extra_shape
                valid_mask = np.zeros(full_shape, dtype=bool)
                valid_mask.reshape(flat_shape)[self._roi] = self._valid_mask
                return valid_mask
            return self._valid_mask.reshape(full_shape)

        return self._valid_mask

    @valid_mask.setter
    def valid_mask(self, valid_nask: np.ndarray):
        """
        :meta private:
        """
        self._valid_mask = valid_nask

    @property
    def valid_slice_bounding(self) -> tuple[slice, ...]:
        """
        Returns a slice that bounds the valid elements in :attr:`data`.

        Note that this includes all valid elements, but also may contain
        invalid elements. If you instead want to get a slice that
        contains no invalid elements (which may cut off some valid elements),
        use :meth:`get_valid_slice_inner`.
        """
        return get_bbox_slice(self.valid_mask)

    def get_valid_slice_inner(self, axis: int = 0) -> tuple[slice, ...]:
        """
        Return a slice into :attr:`data` across `axis` that contains only
        elements marked as valid. This is the "inner" slice, meaning all
        elements selected by the slice are valid according to :attr:`valid_mask`.

        Parameters
        ----------

        axis
            The axis across which we should cut. Other axes are unrestricted in
            the returned slice. For example, if you have a :code:`(y, x)`
            result, specifying :code:`axis=0` will give you a slice like
            :code:`np.s_[y0:y1, :]` where for y in y0..y1, :code:`data[y, :]` is valid
            according to :attr:`valid_mask`.
        """
        return get_inner_slice(self.valid_mask, axis=axis)

    @property
    def masked_data(self) -> np.ma.MaskedArray:
        """
        Same as :attr:`data`, but masked to the valid data, as a numpy
        masked array.
        """
        mask = ~self.valid_mask
        return np.ma.array(self.data, mask=mask)

    @property
    def raw_masked_data(self) -> np.ma.MaskedArray:
        """
        Same as :attr:`raw_data`, but masked to the valid data, as a numpy
        masked array.
        """
        # NOTE: using `self._valid_mask` instead of `self.valid_mask` to get the
        # raw flat mask, not the one expanded to the full
        mask = ~self._valid_mask
        return np.ma.array(self.raw_data, mask=mask)

    @property
    def kind(self) -> Literal['nav', 'sig', 'single']:
        """
        Get the kind of this buffer.

        .. versionadded:: 0.5.0
        """
        return self._kind

    @property
    def extra_shape(self):
        """
        Get the :code:`extra_shape` of this buffer.

        .. versionadded:: 0.5.0
        """
        return self._extra_shape

    @property
    def where(self):
        """
        Get the place where this buffer is to be allocated.

        .. versionadded:: 0.6.0
        """
        return self._where

    def __array__(self):
        """
        returns the "wrapped"/reshaped array, see above
        """
        return self.data

    def allocate(self, lib=None):
        """
        Allocate a new buffer, in the shape previously set
        via one of the `set_shape_*` methods.

        .. versionchanged:: 0.6.0
            Support for allocating on device

        :meta private:
        """
        if self._shape is None:
            raise RuntimeError("cannot allocate: no shape set")
        if self._data is not None:
            raise RuntimeError("cannot allocate: data is already set")
        if self._where == 'device' and lib is not None:
            _z = lib.zeros
        else:
            _z = zeros_aligned
        self._data = _z(self._shape, dtype=self._dtype)

    def has_data(self):
        """
        :meta private:
        """
        return self._data is not None

    @property
    def roi_is_zero(self):
        """
        :meta private:
        """
        return self._roi_is_zero

    def _update_roi_is_zero(self):
        """
        :meta private:
        """
        self._roi_is_zero = prod(self._shape) == 0

    def _slice_for_partition(self, partition):
        """
        Get a Slice into self._data for `partition`, taking the current ROI into account.

        Because _data is "compressed" if a ROI is set, we can't directly index and must
        calculate a new slice from the ROI.

        :meta private:
        """
        if self._roi is None:
            return partition.slice
        return partition.slice.adjust_for_roi(self._roi)

    def get_view_for_dataset(self, dataset):
        """
        :meta private:
        """
        if self._contiguous_cache:
            raise RuntimeError("Cache is not empty, has to be flushed")
        return self._data

    def _get_slice(self, slice: Slice):
        """
        :meta private:
        """
        real_slice = slice._get()
        shape = tuple(slice.shape) + self.extra_shape
        return self._get_slice_direct(real_slice, shape)

    def _get_slice_direct(self, real_slice: slice, shape):
        """
        :meta private:
        """
        result = self._data[real_slice]
        # Defend against #1026 (internal bugs), allow deactivating in
        # optimized builds for performance
        assert result.shape == shape
        return result

    def get_view_for_partition(self, partition):
        """
        get a view for a single partition in a whole-result-sized buffer

        :meta private:
        """
        if self._contiguous_cache:
            raise RuntimeError("Cache is not empty, has to be flushed")
        if self._kind == "nav":
            return self._get_slice(self._slice_for_partition(partition).nav)
        elif self._kind == "sig":
            return self._get_slice(partition.slice.sig)
        elif self._kind == "single":
            return self._data

    def get_view_for_frame(self, partition, tile, frame_idx):
        """
        get a view for a single frame in a partition- or dataset-sized buffer
        (partition-sized here means the reduced result for a whole partition,
        not the partition itself!)

        :meta private:
        """
        if partition.shape.dims != partition.shape.sig.dims + 1:
            raise RuntimeError("partition shape should be flat, is %s" % partition.shape)
        if self._contiguous_cache:
            raise RuntimeError("Cache is not empty, has to be flushed")
        if self._kind == "sig":
            return self._get_slice(tile.tile_slice.sig)
        elif self._kind == "nav":
            partition_slice = self._slice_for_partition(partition)
            if self._data_coords_global:
                offset = 0
            else:
                offset = partition_slice.origin[0]
            # Defense against #1026: if it is an int, there is no empty or out
            # of range slice, but an index error
            result_idx = (int(tile.tile_slice.origin[0] + frame_idx - offset),)
            # shape: (1,) + self._extra_shape
            if len(self._extra_shape) > 0:
                return self._data[result_idx]
            else:
                return self._data[result_idx + (np.newaxis,)]
        elif self._kind == "single":
            return self._data

    def get_view_for_tile(self, partition, tile):
        """
        get a view for a single tile in a partition-sized buffer
        (partition-sized here means the reduced result for a whole partition,
        not the partition itself!)

        :meta private:
        """
        if self._contiguous_cache:
            raise RuntimeError("Cache is not empty, has to be flushed")
        if self.roi_is_zero:
            raise ValueError("cannot get view for tile with zero ROI")
        if self._kind == "sig":
            return self._get_slice(tile.tile_slice.sig)
        elif self._kind == "nav":
            partition_slice = self._slice_for_partition(partition)
            tile_slice = tile.tile_slice
            if self._data_coords_global:
                offset = 0
            else:
                offset = partition_slice.origin[0]
            result_start = tile_slice.origin[0] - offset
            result_stop = result_start + tile_slice.shape[0]
            # Defend against #1026 (internal bugs), allow deactivating in
            # optimized builds for performance
            assert result_start < len(self._data)
            assert result_stop <= len(self._data)
            return self._data[result_start:result_stop]
        elif self._kind == "single":
            return self._data

    def get_contiguous_view_for_tile(self, partition, tile):
        '''
        Make a cached contiguous copy of the view for a single tile
        if necessary.

        Currently this is only necessary for :code:`kind="sig"` buffers.
        Use :meth:`flush` to write back the cache.

        Boundary condition: :code:`tile.tile_slice.get(sig_only=True)`
        does not overlap for different tiles while the cache is active,
        i.e. the tiles follow LiberTEM slicing for
        :meth:`libertem.udf.base.UDFTileMixing.process_tile()`.

        .. versionadded:: 0.5.0

        Returns
        -------

        view : np.ndarray
            View into data or contiguous copy if necessary

        :meta private:
        '''
        if self._kind == "sig":
            key = tile.tile_slice._discard_nav_key()
            if key in self._contiguous_cache:
                view = self._contiguous_cache[key]
            else:
                real_slice, expected_shape = self._slice_from_key(key, self.extra_shape)
                view = self._get_slice_direct(real_slice, expected_shape)
                # See if the signal dimension can be flattened
                # https://docs.scipy.org/doc/numpy/reference/generated/numpy.reshape.html
                if not view.flags.c_contiguous:
                    view = view.copy()
                    self._contiguous_cache[key] = view
            return view
        else:
            return self.get_view_for_tile(partition, tile)

    @staticmethod
    @functools.lru_cache(maxsize=512)
    def _slice_from_key(key, extra_shape):
        """
        :meta private:
        """
        origin, shape, sig_dims = key
        expected_shape = shape[-sig_dims:]
        return (
            tuple(slice(o, (o + s)) for (o, s) in zip(origin[-sig_dims:], expected_shape)),
            expected_shape + extra_shape
        )

    def flush(self, debug=False):
        '''
        Write back any cached contiguous copies

        .. versionadded:: 0.5.0

        :meta private:
        '''
        if self._kind == "sig":
            for key, view in self._contiguous_cache.items():
                sl, _ = self._slice_from_key(key, tuple())
                _, _, sig_dims = key
                sl = sl[-sig_dims:]
                self._data[sl] = view
                # FIXME disjoint() for tuple keys
                # if debug and not disjoint(key, self._contiguous_cache.keys()):
                #     raise RuntimeError(
                #         "`key` %r should be disjoint with existing keys" % key
                #     )
            self._contiguous_cache = dict()
        else:
            if self._contiguous_cache:
                raise RuntimeError(
                    f"Contiguous cache not implemented for kind {self._kind}."
                )

    def export(self):
        '''
        Convert device array to NumPy array for pickling and merging

        :meta private:
        '''
        self._data = to_numpy(self._data)

    def __repr__(self):
        return "<BufferWrapper kind={} dtype={} extra_shape={}>".format(
            self._kind, self._dtype, self._extra_shape
        )

    def replace_array(self, data: "nt.ArrayLike") -> None:
        """
        Set the data backing the BufferWrapper even if the BufferWrapper
        already has self._data allocated

        data should be any array-like object

        Will perform checking for shape and dtype.

        :meta private:
        """
        if self._data is None:
            shape = self._shape
            dtype = self._dtype
        else:
            shape = self._data.shape
            dtype = self._data.dtype
        if data.dtype != dtype:
            raise ValueError(f"dtype mismatch: buffer is {dtype}, data is {data.dtype}.")
        if data.shape != shape:
            raise ValueError(f"Shape mismatch: buffer is {shape}, data is {data.shape}.")
        self._contiguous_cache = dict()
        self._data = data

    def result_buffer_type(self):
        """
        Define the type of Buffer used to return final UDF results

        More specialised buffers can override this

        :meta private:
        """
        return PreallocBufferWrapper


class PlaceholderBufferWrapper(BufferWrapper):
    """
    A declaration-only version of :code:`BufferWrapper` that doesn't
    actually allocate a buffer. Meant as a placeholder for results
    that are only materialized in :code:`UDF.get_results`.
    """
    def allocate(self, lib=None):
        return None

    def has_data(self):
        return False

    def export(self):
        return None

    def get_view_for_partition(self, partition):
        return None

    def get_view_for_frame(self, partition, tile, frame_idx):
        return None

    def get_view_for_tile(self, partition, tile):
        return None

    def get_contiguous_view_for_tile(self, partition, tile):
        return None

    @property
    def data(self):
        raise ValueError(
            "this BufferWrapper doesn't have a value associated with it"
        )

    @property
    def raw_data(self):
        raise ValueError(
            "this BufferWrapper doesn't have a value associated with it"
        )


class PreallocBufferWrapper(BufferWrapper):
    def __init__(self, data, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._data = data


class AuxBufferWrapper(BufferWrapper):
    def new_for_partition(self, partition, roi):
        """
        Return a new AuxBufferWrapper for a specific partition,
        slicing the data accordingly and reducing it to the selected roi.

        This assumes to be called on an AuxBufferWrapper that was not created
        by this method, that is, it should have global coordinates without
        having the ROI applied.
        """
        # FIXME: right now creates a view for the partition slice, which
        # AFAIK means we serialize the whole array; we could optimize here
        # and only send over the partition slice. But maybe, later, when we
        # actually properly scatter and share data, this becomes obsolete anyways,
        # as we would scatter most likely for all partitions (to be flexible in node
        # assignment, for example for availability)
        assert self._data_coords_global
        ps = partition.slice.get(nav_only=True)
        buf = self.__class__(self._kind, self._extra_shape, self._dtype)
        if roi is not None:
            roi_part = roi.reshape(-1)[ps]
            new_data = self._data[ps][roi_part]
        else:
            new_data = self._data[ps]
        buf.set_buffer(new_data, is_global=False)
        buf.set_roi(roi)
        assert prod(new_data.shape) > 0
        assert not buf._data_coords_global
        return buf

    def get_view_for_dataset(self, dataset):
        return self._data[self._roi]

    def set_buffer(self, buf, is_global=True):
        """
        Set the underlying buffer to an existing numpy array.

        If is_global is True, the shape must match with the shape of nav or sig
        of the dataset, plus extra_shape, as determined by the `kind` and
        `extra_shape` constructor arguments.
        """
        assert self._data is None
        assert buf.dtype == self._dtype
        extra = self._extra_shape
        shape = (-1,)
        if extra and extra != (1,):
            shape = shape + extra
        self._data = buf.reshape(shape)
        self._data_coords_global = is_global

    def __repr__(self):
        return "<AuxBufferWrapper kind={} dtype={} extra_shape={}>".format(
            self._kind, self._dtype, self._extra_shape
        )



import os
import platform
from typing import Optional


'''
Get and set environment variables that indicate to use CPU or GPU processing and
the appropriate device IDs

This allows configuring a Dask worker or inline executor to use CPU or a
particular GPU, which allows simultaneous use of all GPUs and CPUs on a
processing node without oversubscription.

Internally, this gets and sets the environment variables
:code:`LIBERTEM_USE_CUDA` and :code:`LIBERTEM_USE_CPU`.

See :ref:`udf cuda` for details on how to run UDFs with CUDA support and
:ref:`debugging udfs` for information on how to use these functions to control
where an :class:`~libertem.executor.inline.InlineJobExecutor` runs tasks.
'''


def get_use_cuda() -> Optional[int]:
    '''
    .. versionadded:: 0.6.0

    Returns
    -------
    id : int or None
        CUDA device ID to use.
    '''
    ret = os.environ.get("LIBERTEM_USE_CUDA")
    if ret is not None:
        return int(ret)
    else:
        return None


def set_use_cuda(cuda_device: int):
    '''
    This sets a CUDA device ID and unsets any CPU ID

    .. versionadded:: 0.6.0

    Parameters
    ----------
    cuda_device : int
        CUDA device ID to use
    '''
    os.environ["LIBERTEM_USE_CUDA"] = str(cuda_device)
    os.environ.pop("LIBERTEM_USE_CPU", None)


def get_use_cpu():
    '''
    .. versionadded:: 0.6.0

    Returns
    -------
    id : int or None
        CPU device ID to use. Currently there is no pinning, i.e. the value itself is
        ignored. :code:`None` means "don't use CPU" and any integer means "use CPU".
        Default is 0 if no settings were applied before.
    '''
    ret = os.environ.get("LIBERTEM_USE_CPU")
    if ret is not None:
        ret = int(ret)
    elif get_use_cuda() is None:
        # If no variable is set, return CPU 0
        # For example inline executor or test code
        ret = 0
    return ret


def set_use_cpu(cpu: int):
    '''
    This sets a CPU device ID and unsets any CUDA ID

    .. versionadded:: 0.6.0

    Parameters
    ----------
    cpu : int
        CPU to use. The value is currently ignored, i.e. any CPU is used without pinning
    '''
    os.environ.pop("LIBERTEM_USE_CUDA", None)
    os.environ["LIBERTEM_USE_CPU"] = str(cpu)


def get_device_class():
    '''
    .. versionadded:: 0.6.0

    Returns
    -------
    class : str
        Device class to use. Can be 'cpu' or 'cuda'. Default is 'cpu' if
        no settings were applied before.
    '''
    cuda = get_use_cuda()
    cpu = get_use_cpu()
    if cpu is not None and cuda is not None:
        raise RuntimeError(
            "Both LIBERTEM_USE_CPU and LIBERTEM_USE_CUDA set, expecting at most one"
        )
    if cuda is not None:
        return "cuda"
    else:
        return "cpu"


def set_file_limit():
    if platform.system() == "Windows":
        # Windows has no problem opening huge numbers of files
        return

    import resource
    _, hard_lim = resource.getrlimit(resource.RLIMIT_NOFILE)
    resource.setrlimit(resource.RLIMIT_NOFILE, (hard_lim, hard_lim))



import numpy as np
import pytest

from libertem.udf import UDF
from libertem.io.dataset.memory import MemoryDataSet

from utils import _mk_random


class PixelsumUDF(UDF):
    def get_result_buffers(self):
        return {
            'pixelsum': self.buffer(
                kind="nav", dtype="float32"
            )
        }

    def process_partition(self, partition):
        self.results.pixelsum[:] += np.sum(partition, axis=(-1, -2))


@pytest.mark.parametrize(
    'tileshape', (None, ((16*16)//7, 16, 16))
)
def test_sum_tiles(lt_ctx, tileshape):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    # The dataset can force a different tile size even for process_partition!
    # At least MemoryDataSet does that and we should check that everything is ok...
    dataset = MemoryDataSet(
        data=data, num_partitions=7, sig_dims=2, tileshape=tileshape
    )

    pixelsum = PixelsumUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=pixelsum)
    assert 'pixelsum' in res
    print(data.shape, res['pixelsum'].data.shape)
    assert np.allclose(res['pixelsum'].data, np.sum(data, axis=(2, 3)))


class TouchUDF(UDF):
    def get_result_buffers(self):
        return {
            'touched': self.buffer(
                kind="nav", dtype="int"
            )
        }

    def process_partition(self, partition):
        print(partition.shape)
        self.results.touched[:] += 1
        assert partition.shape[0] == self.meta.coordinates.shape[0]


@pytest.mark.parametrize(
    'use_roi', (False, True)
)
@pytest.mark.parametrize(
    'tileshape', (None, (7, 16, 16), (8 * 16, 16, 16))
)
def test_partition_roi(lt_ctx, use_roi, tileshape):
    data = _mk_random(size=(16, 16, 16, 16), dtype="float32")
    dataset = MemoryDataSet(data=data, num_partitions=2, sig_dims=2, tileshape=tileshape)
    if use_roi:
        roi = np.random.choice([True, False], dataset.shape.nav)
    else:
        roi = None
    udf = TouchUDF()
    res = lt_ctx.run_udf(dataset=dataset, udf=udf, roi=roi)
    print(data.shape, res['touched'].data.shape)
    assert np.all(res['touched'].raw_data == 1)



import itertools

import pytest
import numpy as np

from libertem.io.dataset.base.decode import (
    default_decode, decode_swap_2, decode_swap_4, decode_swap_8,
    DtypeConversionDecoder,
)


@pytest.mark.with_numba
def test_default_decode():
    inp = np.random.randn(1, 16, 16).astype(np.float32)
    out = np.zeros_like(inp, dtype=np.float64)
    rr = np.array([0, 0, inp.nbytes])
    origin = np.array([0, 0, 0])
    shape = np.array([1, 16, 16])
    ds_shape = np.array([256, 16, 16])
    default_decode(
        inp.reshape((-1,)).view(dtype=np.uint8),
        out=out.reshape((1, -1,)),
        idx=0,
        native_dtype=inp.dtype,
        rr=rr,
        origin=origin,
        shape=shape,
        ds_shape=ds_shape,
    )
    assert np.allclose(
        inp.astype(np.float64), out
    )


def hexdump(arr):
    arr = arr.reshape((-1,))
    return " ".join("%02x" % i for i in arr.view(dtype=np.uint8))


@pytest.mark.xfail(reason="not implemented yet")
@pytest.mark.with_numba
def test_decode_swap_4_f32():
    inp = np.random.randn(1, 16, 16).astype(np.float32)
    out = np.zeros_like(inp, dtype=inp.dtype)
    rr = np.array([0, 0, inp.nbytes])
    origin = np.array([0, 0, 0])
    shape = np.array([1, 16, 16])
    ds_shape = np.array([256, 16, 16])
    inp_swapped = inp.reshape((-1,)).byteswap()
    decode_swap_4(
        inp=inp_swapped.view(dtype=np.uint8),
        out=out.reshape((1, -1)),
        idx=0,
        native_dtype=inp.dtype,
        rr=rr,
        origin=origin,
        shape=shape,
        ds_shape=ds_shape,
    )
    print(hexdump(inp_swapped[:2]))
    print(hexdump(inp.reshape((-1,))[:2]))
    print(hexdump(out.reshape((-1,))[:2]))
    assert np.allclose(inp, out)


@pytest.mark.xfail(reason="not implemented yet")
@pytest.mark.with_numba
def test_decode_swap_4_f64():
    inp = np.random.randn(1, 16, 16).astype(np.float64)
    out = np.zeros_like(inp, dtype=inp.dtype)
    rr = np.array([0, 0, inp.nbytes])
    origin = np.array([0, 0, 0])
    shape = np.array([1, 16, 16])
    ds_shape = np.array([256, 16, 16])
    inp_swapped = inp.reshape((-1,)).byteswap()
    decode_swap_8(
        inp=inp_swapped.view(dtype=np.uint8),
        out=out.reshape((1, -1)),
        idx=0,
        native_dtype=inp.dtype,
        rr=rr,
        origin=origin,
        shape=shape,
        ds_shape=ds_shape,
    )
    print(hexdump(inp_swapped[:2]))
    print(hexdump(inp.reshape((-1,))[:2]))
    print(hexdump(out.reshape((-1,))[:2]))
    assert np.allclose(inp, out)


@pytest.mark.with_numba
def test_decode_swap_2_u16():
    inp = np.random.randint(low=0, high=2**15, size=(1, 16, 16), dtype=np.uint16)
    out = np.zeros_like(inp, dtype=inp.dtype)
    rr = np.array([0, 0, inp.nbytes])
    origin = np.array([0, 0, 0])
    shape = np.array([1, 16, 16])
    ds_shape = np.array([256, 16, 16])
    inp_swapped = inp.reshape((-1,)).byteswap()
    decode_swap_2(
        inp=inp_swapped.view(dtype=np.uint8),
        out=out.reshape((1, -1)),
        idx=0,
        native_dtype=inp.dtype,
        rr=rr,
        origin=origin,
        shape=shape,
        ds_shape=ds_shape,
    )
    print(hexdump(inp_swapped[:1]))
    print(hexdump(inp.reshape((-1,))[:1]))
    print(hexdump(out.reshape((-1,))[:1]))
    assert np.allclose(inp, out)


@pytest.mark.with_numba
def test_decode_swap_4_u32():
    inp = np.random.randint(low=0, high=2**31, size=(1, 16, 16), dtype=np.uint32)
    out = np.zeros_like(inp, dtype=inp.dtype)
    rr = np.array([0, 0, inp.nbytes])
    origin = np.array([0, 0, 0])
    shape = np.array([1, 16, 16])
    ds_shape = np.array([256, 16, 16])
    inp_swapped = inp.reshape((-1,)).byteswap()
    decode_swap_4(
        inp=inp_swapped.view(dtype=np.uint8),
        out=out.reshape((1, -1)),
        idx=0,
        native_dtype=inp.dtype,
        rr=rr,
        origin=origin,
        shape=shape,
        ds_shape=ds_shape,
    )
    print(hexdump(inp_swapped[:1]))
    print(hexdump(inp.reshape((-1,))[:1]))
    print(hexdump(out.reshape((-1,))[:1]))
    assert np.allclose(inp, out)


@pytest.mark.with_numba
def test_decode_swap_4_u64():
    inp = np.random.randint(low=0, high=2**31, size=(1, 16, 16), dtype=np.uint64)
    out = np.zeros_like(inp, dtype=inp.dtype)
    rr = np.array([0, 0, inp.nbytes])
    origin = np.array([0, 0, 0])
    shape = np.array([1, 16, 16])
    ds_shape = np.array([256, 16, 16])
    inp_swapped = inp.reshape((-1,)).byteswap()
    decode_swap_8(
        inp=inp_swapped.view(dtype=np.uint8),
        out=out.reshape((1, -1)),
        idx=0,
        native_dtype=inp.dtype,
        rr=rr,
        origin=origin,
        shape=shape,
        ds_shape=ds_shape,
    )
    print(hexdump(inp_swapped[:2]))
    print(hexdump(inp.reshape((-1,))[:2]))
    print(hexdump(out.reshape((-1,))[:2]))
    assert np.allclose(inp, out)


@pytest.mark.parametrize(
    'dtypes, in_byteorder, out_byteorder',
    itertools.product(
        [
            (np.uint8, np.uint16),
            (np.uint8, np.uint32),
            (np.uint8, np.uint64),
            (np.uint16, np.uint16),
            (np.uint16, np.uint32),
            (np.uint16, np.uint64),
            (np.uint32, np.uint32),
            (np.uint32, np.uint64),

            (np.uint8, np.float32),
            (np.uint16, np.float32),
            (np.uint32, np.float32),

            (np.uint8, np.float64),
            (np.uint16, np.float64),
            (np.uint32, np.float64),
        ],
        ['<', '>', '='],  # input byteorder
        ['='],  # decoding to non-native byteorder is not supported currently
    ),
)
@pytest.mark.with_numba
def test_default_decoder_from_uint(dtypes: tuple[np.dtype], in_byteorder, out_byteorder):
    in_dtype, out_dtype = dtypes
    decoder = DtypeConversionDecoder()
    in_dtype = np.dtype(in_dtype)
    out_dtype = np.dtype(out_dtype)
    in_dtype_full = in_dtype.newbyteorder(in_byteorder)
    out_dtype_full = out_dtype.newbyteorder(out_byteorder)
    decode = decoder.get_decode(native_dtype=in_dtype_full, read_dtype=out_dtype_full)
    inp = np.random.randint(
        low=0,
        high=np.iinfo(in_dtype).max,
        size=(1, 16, 16),
        dtype=in_dtype
    )
    if in_dtype_full != in_dtype:
        inp.byteswap()
    print(in_dtype_full, in_dtype, out_dtype_full, out_dtype)
    out = np.zeros_like(inp, dtype=out_dtype_full)
    rr = np.array([0, 0, inp.nbytes])
    origin = np.array([0, 0, 0])
    shape = np.array([1, 16, 16])
    ds_shape = np.array([256, 16, 16])

    if decoder._need_byteswap(in_dtype_full, out_dtype_full):
        assert decoder.get_native_dtype(in_dtype_full, out_dtype_full) == np.uint8
        swapping = True
    else:
        assert decoder.get_native_dtype(in_dtype_full, out_dtype_full) == in_dtype_full
        swapping = False
    decode(
        inp=inp.reshape((-1,)).view(dtype=np.uint8),
        out=out.reshape((1, -1)),
        idx=0,
        native_dtype=inp.dtype,
        rr=rr,
        origin=origin,
        shape=shape,
        ds_shape=ds_shape,
    )

    # swap back and compare:
    if swapping:
        expected = inp.byteswap()
    else:
        expected = inp
    expected = expected.astype(out_dtype_full)

    print(swapping)
    print(hexdump(inp.reshape((-1,))[:1]))
    print(hexdump(expected.reshape((-1,))[:1]))
    print(hexdump(out.reshape((-1,))[:1]))

    assert np.allclose(expected, out)
